(function() {
	'use strict';

	angular.module('ipconfigure.orchid.core').factory('IdleService', ['$document', '$rootScope', '$timeout', IdleService]);

	function IdleService($document, $rootScope, $timeout) {
		var _idleTickCount = -1,
			_idleSubscriptions = [];

		function _fireCallbacks(seconds, secondsExceeded) {
			var i;

			if (Array.isArray(_idleSubscriptions[seconds]) === true) {
				for (i = 0; i < _idleSubscriptions[seconds].length; i += 1) {
					if (angular.isUndefined(secondsExceeded) === true || _idleSubscriptions[seconds][i].seconds <= secondsExceeded) {
						_idleSubscriptions[seconds][i].callback();
					}
				}
			}

			$rootScope.$apply();
		}

		function _onDocumentActivity() {
			var i,
				previousIdleTickCount = _idleTickCount;

			_idleTickCount = 0;

			// only fire this if someone is expecting it (meaning we've broadcast another message prior to this)
			_fireCallbacks(0, previousIdleTickCount);
		}

		function _createSubscription(seconds, subscription) {
			if (Array.isArray(_idleSubscriptions[seconds]) === false) {
				_idleSubscriptions[seconds] = [];
			}

			_idleSubscriptions[seconds].push(subscription);
		}

		function _destroySubscription(seconds, key) {
			[0, seconds].forEach(function(which) {
				var i;

				if (Array.isArray(_idleSubscriptions[which]) === true) {
					for (i = 0; i < _idleSubscriptions[which].length; i += 1) {
						if (_idleSubscriptions[which][i].key === key) {
							_idleSubscriptions[which].splice(i, 1);
							i--;
						}
					}
				}
			});
		}

		function _init() {
			// monitor for mousemove so we know when to chill and when to wile out
			$document.bind('keydown mousedown mousemove mousewheel DOMMouseScroll click TouchStart', _onDocumentActivity);

			// ~infinite loop checking for inactivity
			(function loop() {
				var i;

				_idleTickCount++;

				// fire event if a subscription matches the current idleTickCount (but not the first time, because we dont want to needlessly fire _onUserAction
				if (_idleTickCount > 0) {
					_fireCallbacks(_idleTickCount);
				}

				// invokeApply=false for performance gains
				$timeout(loop, 1000, false);
			}());
		}

		_init();

		function _monitorIdleTime(seconds) {
			var key = Math.random(),
				registration;

			registration = {
				onIdle: function(callback) {
					_createSubscription(seconds, {
						seconds: seconds,
						key: key,
						callback: callback
					});

					return registration;
				},
				onReturnFromIdle: function(callback) {
					_createSubscription(0, {
						seconds: seconds,
						key: key,
						callback: callback
					});

					return registration;
				},
				cancel: function() {
					_destroySubscription(seconds, key);
				}
			};

			return registration;
		}

		return {
			monitorIdleTime: _monitorIdleTime
		};
	}
}());
