(function() {
	'use strict';

	angular.module('ipconfigure.orchid').controller('AdminActivationCtrl', ['$scope', '$filter', '$http', '$log', '$timeout', 'BaseServiceUrl', 'OrchidDiscoveryService', AdminActivationCtrl]);

	function AdminActivationCtrl($scope, $filter, $http, $log, $timeout, BaseServiceUrl, OrchidDiscoveryService) {
		$scope.currentInstance = OrchidDiscoveryService.currentInstance;
		$scope.licenseSession = {};
		$scope.processing = false;
		$scope.processingDeactivate = false;
		$scope.failure = '';
		$scope.failureDeactivate = '';
		$scope.override = false;
		$scope.deactivateStep = 1;
		$scope.detachStep = 1;
		$scope.trustedIssuer = undefined;

		//$scope.href = 'data:text/plain;charset=utf-8,' + encodeURIComponent($scope.activation.machineKey);

		$scope.$watch('currentInstance', function() {
			$scope.licenseSession = angular.copy($scope.currentInstance.licenseSession);

			// fill in properties
			if (typeof $scope.licenseSession.license.retention === 'number') {
				$scope.licenseSession.license.retentionDescription = $filter('bytes')($scope.licenseSession.license.retention);
			}
			else {
				$scope.licenseSession.license.retentionDescription = 'Unlimited';
			}

			if ($scope.licenseSession.license.expires > 8623372036854776) {
				$scope.licenseSession.license.expiresDescription = 'N/A';
			}
			else {
				$scope.licenseSession.license.expiresDescription = $filter('date')($scope.licenseSession.license.expires, 'fullDate', 'UTC');
			}

			if (angular.isUndefined($scope.licenseSession.license.activationCode) || $scope.licenseSession.license.activationCode.length === 0) {
				$scope.licenseSession.license.activationCode = 'N/A';
			}
		}, true);

		$scope.activate = function() {
			$scope.failure = '';
			$scope.processing = true;

			$http.post(BaseServiceUrl + '/service/license-session', $scope.activationKey)
				.success(function(response) {
					$timeout(function() {
						// update the service with the new license information
						OrchidDiscoveryService.updateLicenseSession(response);

						$scope.failure = '';
						$scope.processing = false;
						$scope.override = false;
						$scope.activationKey = '';
					}, 2500);
				})
				.error(function(data, status, headers, config) {
					$scope.processing = false;

					if (data && (data.code || data.reason)) {
						$scope.failure = data.code + ': ' + data.reason;
					}
					else {
						$scope.failure = 'An unexpected error occurred.';
					}

					if (status !== 0) {
						$log.error('error, POST /service/license', data, status, config);
					}
				});
		};

		$scope.deactivate = function() {
			if ($scope.deactivateStep === 3) {
				$scope.processingDeactivate = true;
				$http.delete(BaseServiceUrl + '/service/license-session', $scope.activationKey)
					.success(function(response) {
						// update the service with the new license information
						OrchidDiscoveryService.updateLicenseSession(response);

						$scope.processingDeactivate = false;
						$scope.override = false;
						$scope.activationKey = '';
						//$scope.$apply();
					})
					.error(function(data, status, headers, config) {
						$scope.processingDeactivate = false;
						$scope.deactivateStep = 1;

						if (data && (data.code || data.reason)) {
							$scope.failureDeactivate = data.code + ': ' + data.reason;
						}
						else {
							$scope.failureDeactivate = 'An unexpected error occurred.';
						}

						if (status !== 0) {
							$log.error('error, POST /service/license', data, status, config);
						}
					});
			}
			else {
				$scope.deactivateStep += 1;
			}
		};

		$scope.detach = function() {	
			if ($scope.detachStep == 4) {
				var ret = confirm("After manually detaching your Orchid from Fusion, you will still need to remove Orchid from Fusion.\n\nDo you understand this?");
				if (ret == true) {
					$scope.processingDetach = true;
					$http.delete(BaseServiceUrl + '/service/trusted/issuer')
						.success(function(response) {
							$scope.trustedIssuer = undefined;
							$scope.processingDetach = false;
						})
						.error(function(data, status, headers, config) {
							$scope.processingDetach = false;
							$scope.detachStep = 1;
							$scope.failureDetach = "An unexpected error occurred.";
						})
				}
			} else {
				$scope.detachStep += 1;
			}
		};

		$scope.overrideLicenseView = function() {
			$scope.override = true;
			$scope.deactivateStep = 1;
		};

		function _init() {
			$http.get(BaseServiceUrl + '/service/trusted/issuer')
				.success(function(response) {
					$scope.trustedIssuer = response;
				})
				.error(function(data, status, headers, config) {
					// No fusion attached
				});
		}

		_init();

		//$scope.downloadKey = function () {
		//	// not angularish
		//	document.querySelector('#download').click();
		//};
	}
}());
