(function() {
	'use strict';

	angular.module('ipconfigure.orchid').controller('AdminCameraAddCtrl', ['$scope', '$http', '$log', '$timeout', 'FocusOnService', 'OrchidDiscoveryService', 'ScrollToOnService', AdminCameraAddCtrl]);

	function AdminCameraAddCtrl($scope, $http, $log, $timeout, FocusOnService, OrchidDiscoveryService, ScrollToOnService) {
		var defaultOnvifUrl = 'http://192.168.1.1/onvif/device_service',
			defaultGenericUrl = 'rtsp://192.168.1.1/axis-media/media.amp';

		$scope.instance = OrchidDiscoveryService.currentInstance;
		$scope.orchidInstances = OrchidDiscoveryService.instances;

		$scope.registration = {
			driver: 'ONVIF',
			name: '',
			import: false,
			connection: {
				uri: defaultOnvifUrl,
				username: '',
				password: ''
			}
		};

		$scope.$watch('registration.driver', function(newValue) {
			if ($scope.registration.connection.uri === defaultOnvifUrl || $scope.registration.connection.uri === defaultGenericUrl) {
				switch (newValue) {
					case 'ONVIF':
						$scope.registration.connection.uri = defaultOnvifUrl;
						break;
					case 'Generic RTSP':
						$scope.registration.connection.uri = defaultGenericUrl;
						break;
				}
			}
		});

		$scope.registerCamera = function() {
			$scope.processing = true;

			$http.post($scope.instance.baseUrl + '/service/cameras', $scope.registration)
				.success(function(response) {
					$log.debug('registration complete', response);

					response.instance = $scope.instance;

					// not entirely sure about this -- we're essentially modifying the parent scope
					$scope.configuredCameras.push(response);
					$scope.goCameraEdit(response);

					$scope.failure = '';
					$scope.processing = false;
				})
				.error(function(data, status, headers, config) {
					if (data && (data.code || data.reason)) {
						$scope.failure = data.code + ': ' + data.reason;
					}
					else {
						$scope.failure = 'An unexpected error occurred.';
					}

					$scope.processing = false;

					if (status !== 0) {
						$log.error('error, POST /service/cameras', data, status, config);
					}
				});
		};

		ScrollToOnService('top');

		// allow for scroll to happen
		$timeout(function() {
			FocusOnService('initial-field');
		}, 275);
	}
}());
