(function() {
	'use strict';

	angular.module('ipconfigure.orchid').controller('AdminCameraBulkRegistrationCtrl', ['$scope', '$filter', '$http', '$log', '$timeout', 'FocusOnService', 'OrchidDiscoveryService', AdminCameraBulkRegistrationCtrl]);

	function AdminCameraBulkRegistrationCtrl($scope, $filter, $http, $log, $timeout, FocusOnService, OrchidDiscoveryService) {
		function _registerCamera(baseUrl, camera) {
			var registration = {
				driver: 'ONVIF',
				connection: {
					uri: camera.url,
					username: $scope.username,
					password: $scope.password
				},
				name: camera.ip
			};

			return $http.post(baseUrl + '/service/cameras', registration);
		}

		function _refreshDiscoveredCameras() {
			$scope.discoveredCamerasCopy = angular.copy($scope.discoveredCameras);

			// provide existing registration information
			$scope.discoveredCamerasCopy.forEach(function(discoveredCamera) {
				if ($scope.bulkChecked === true) {
					$scope.selectCamera(discoveredCamera);
					$scope.ready = true; // We are ready provided at least one camera has been selected
				}
				discoveredCamera.configured = $scope.configuredCameras.some(function(configuredCamera) {
					if (angular.isUndefined(configuredCamera.configuration.IPAddress) === false) {
						return discoveredCamera.ip === configuredCamera.configuration.IPAddress.trim();
					}
				});
			});
		}

		$scope.refresh = function() {
			if ($scope.processing !== true && $scope.processingRefresh !== true) {
				$scope.processingRefresh = true;
				$scope.ready = false;
				$timeout(function() {
					_refreshDiscoveredCameras();
					$scope.processingRefresh = false;
					$scope.registrationStatus.show = false;
				}, 333);
			}
		};

		$scope.toggle = function($event) {
			$scope.bulkChecked = !$scope.bulkChecked;
			var checked = $event.currentTarget.checked;
			var options = document.getElementsByClassName("bulk-option");
			angular.forEach($scope.discoveredCamerasCopy, function(camera) {
				if (camera.selected !== checked) {
					$scope.selectCamera(camera);
				}
			});
		}

		$scope.ipOrder = function(camera) {
			var index = 3,
			result = 0,
			ipParts = camera.ip.split('.');

			// convert the IP address into its numberical representation
			ipParts.forEach(function(ipPart) {
				result += Number(ipPart) * Math.pow(1000, index--);
			});
			return ($scope.ipOrderDir === "+") ? result : -1 * result;
		};

		$scope.orderBy = function(byWhat, $event) {
			$event.preventDefault();

			if ($scope.processing !== true) {
				switch (byWhat) {
					case 'name':
						$scope.ipOrderDir = undefined;
						if ($scope.preferredOrder[0] === '+name') {
							$scope.preferredOrder[0] = '-name';
						}
						else {
							$scope.preferredOrder = ['+name', 'ip'];
						}
						break;
					case 'ip':
						$scope.preferredOrder = [$scope.ipOrder, 'name'];
						if ($scope.ipOrderDir === "+") {
							$scope.ipOrderDir = "-";
						}
						else {
							$scope.ipOrderDir = "+";
						}
						break;
				}
			}
		};

		$scope.selectCamera = function(camera) {
			if ($scope.processing !== true) {
				camera.selected = !camera.selected;

				$scope.ready = $scope.discoveredCamerasCopy.some(function(discoveredCamera) { return discoveredCamera.selected; });
			}
		};

		$scope.showError = function(camera) {
			alert(camera.error);
		};

		$scope.registerCameras = function() {
			var index = 0;

			// initialize registration bits
			$scope.processing = true;
			$scope.ready = false;
			$scope.registrationStatus = {
				show: true,
				errors: [],
				progress: 0,
				progressPercentage: 0,
				total: 0,
				complete: false,
				type: 'success'
			};
			$scope.bulkChecked = false;

			// sort cameras so we can register them in order according to the UI
			$scope.discoveredCamerasCopy = $filter('orderBy')($scope.discoveredCamerasCopy, $scope.preferredOrder);

			// set up cameras for registering
			$scope.discoveredCamerasCopy.forEach(function(camera) {
				delete camera.error;
				delete camera.success;

				if (camera.selected === true) {
					$scope.registrationStatus.total += 1;

					camera.chosen = true;
					camera.selected = false;
				}
				else {
					camera.chosen = false;
					camera.hidden = true;
				}
			});

			(function go() {
				if ($scope.discoveredCamerasCopy.length > index) {
					if ($scope.discoveredCamerasCopy[index].chosen === true) {
						$scope.registrationStatus.progress += 1;
						$scope.registrationStatus.progressPercentage = $scope.registrationStatus.progress / $scope.registrationStatus.total * 100;
						$scope.discoveredCamerasCopy[index].processing = true;

						_registerCamera(OrchidDiscoveryService.currentInstance.baseUrl, $scope.discoveredCamerasCopy[index])
							.success(function(camera) {
								$scope.discoveredCamerasCopy[index].processing = false;
								$scope.discoveredCamerasCopy[index].success = true;

								// set the orchid instance where the camera was registered
								camera.instance = OrchidDiscoveryService.currentInstance;

								// add the camera to the list of configuredCameras (parent scope)
								$scope.configuredCameras.push(camera);

								index += 1;

								if ($scope.destroyed === false) {
									go();
								}
							})
							.error(function(data, status, headers, config) {
								$scope.discoveredCamerasCopy[index].processing = false;
								$scope.discoveredCamerasCopy[index].success = false;

								if (data && (data.code || data.reason)) {
									$scope.discoveredCamerasCopy[index].error = data.code + ': ' + data.reason;
								}
								else {
									$scope.discoveredCamerasCopy[index].error = 'An unexpected error occurred.';
								}

								$scope.registrationStatus.errors.push($scope.discoveredCamerasCopy[index].error);

								// turn the progress bar to a warning
								$scope.registrationStatus.type = 'danger';

								index += 1;

								if (status !== 0) {
									$log.error('error, POST /service/cameras', data, status, config);
								}

								if ($scope.destroyed === false) {
									go();
								}
							});
					}
					else {
						index += 1;

						if ($scope.destroyed === false) {
							go();
						}
					}
				}
				else {
					// we're done
					$scope.processing = false;
					$scope.registrationStatus.complete = true;
				}
			}());
		};

		$scope.removeSuccessfullyRegisteredCameras = function() {
			var ready = false;

			$scope.discoveredCamerasCopy.forEach(function(camera) {
				if (camera.chosen === true) {
					if (camera.success === false) {
						camera.selected = true;

						ready = true;
					}
					else {
						camera.hidden = true;
						camera.chosen = false;
						camera.selected = false;
					}
				}

				$scope.ready = ready;
			});
		};

		$scope.$on('$destroy', function() {
			// cancel the registration loop
			$scope.destroyed = true;
		});

		// clone discoveredCameras
		_refreshDiscoveredCameras();

		// defaults
		$scope.destroyed = false;
		$scope.ready = false;
		$scope.username = '';
		$scope.password = '';
		$scope.preferredOrder = $scope.ipOrder;
		$scope.ipOrderDir = '+';
		$scope.registrationStatus = {
			show: false
		};

		FocusOnService('initial-field');
	}
}());
