(function() {
	'use strict';

	angular.module('ipconfigure.orchid').controller('AdminCameraRetentionPolicyCtrl', ['$scope', '$filter', '$http', '$log', '$q', '$timeout', 'BaseServiceUrl', 'CameraService', 'FocusOnService', 'OrchidDiscoveryService', AdminCameraRetentionPolicyCtrl]);

	function AdminCameraRetentionPolicyCtrl($scope, $filter, $http, $log, $q, $timeout, BaseServiceUrl, CameraService, FocusOnService, OrchidDiscoveryService) {
		var promises = [];

		function fetchConfiguredCameras() {
			var configuredCamerasPromiseTimer,
				deferred = $q.defer(),
				unwatch;

			$scope.cameraServiceState = CameraService.state;

			// wait for CameraService cameras and fulfill the promise
			unwatch = $scope.$watch('cameraServiceState.currentState', function() {
				if ($scope.cameraServiceState.currentState === 'loaded') {
					unwatch();

					$timeout.cancel(configuredCamerasPromiseTimer);

					deferred.resolve(CameraService.cameras);
				}
			});

			// handle the $watch never getting called
			configuredCamerasPromiseTimer = $timeout(function() {
				unwatch();

				deferred.reject('CameraService timed out');
			}, 5000);

			return deferred.promise;
		}

		function fetchDiskUtes() {
			return $http.get(BaseServiceUrl + '/service/server/utilization/disk');
		}

		function fetchRetention() {
			return $http.get(BaseServiceUrl + '/service/cameras/disk-usage');
		}

		function initializeDiskUtes(response) {
			$scope.diskUtes = response.storage;

			$scope.diskUtes.availableUpToThreshold = response.cleanerThreshold;
			$scope.diskUtes.unavailablePercentage = response.storage.unavailable / response.storage.capacity * 100;
			$scope.diskUtes.thresholdPercentage = (response.storage.unavailable + response.cleanerThreshold) / response.storage.capacity * 100;

			// to be processed later
			$scope.diskUtes.bytesFromDeletedCameras = 0;
			$scope.diskUtes.bytesFromDeletedCamerasPercentage = 0;
		}

		function initializeCameraMeta() {
			$scope.configuredCameras.forEach(function(configuredCamera, index) {
				$scope.configuredCamerasMeta[configuredCamera.id] = {
					bytesEstimated: 0,
					bytesEstimatedPercentage: 0,
					bytesEstimatedPerDay: 0,
					bytesRecorded: 0,
					bytesRecordedPercentage: 0,
					daysEstimated: 0,
					secondsRecorded: 0,
					color: $scope.colors[index],
					daysDesired: (configuredCamera.retention / 86400000),
					id: configuredCamera.id,
					override: (configuredCamera.retention > 0)
				};
			});
		}

		function initializeCameraRetention(retention) {
			if (retention && Array.isArray(retention.cameras) === true) {
				retention.cameras.forEach(function(cameraUsage) {
					var cameraMeta = $scope.configuredCamerasMeta[cameraUsage.camera.id];

					if (angular.isUndefined(cameraMeta) === false) {
						cameraMeta.bytesRecorded = cameraUsage.bytesRecorded;
						cameraMeta.secondsRecorded = cameraUsage.secondsRecorded;

						cameraMeta.bytesRecordedPercentage = cameraMeta.bytesRecorded / $scope.diskUtes.capacity * 100;

						// calculate bytesPerDay based on the largest value we have
						if (cameraMeta.secondsRecorded >= 60 * 60 * 24) {
							cameraMeta.bytesEstimatedPerDay = cameraMeta.bytesRecorded / (cameraMeta.secondsRecorded / 60 / 60 / 24);
						}
						else if (cameraMeta.secondsRecorded >= 60 * 60) {
							cameraMeta.bytesEstimatedPerDay = (cameraMeta.bytesRecorded / (cameraMeta.secondsRecorded / 60 / 60)) * 24;
						}
						else if (cameraMeta.secondsRecorded >= 60) {
							cameraMeta.bytesEstimatedPerDay = (cameraMeta.bytesRecorded / (cameraMeta.secondsRecorded / 60)) * 60 * 24;
						}
						else {
							cameraMeta.bytesEstimatedPerDay = 0;
						}
					}
					else {
						$scope.diskUtes.bytesFromDeletedCameras += cameraUsage.bytesRecorded;
					}
				});

				$scope.diskUtes.bytesFromDeletedCamerasPercentage = $scope.diskUtes.bytesFromDeletedCameras / $scope.diskUtes.capacity * 100;
			}
		}

		function calculateCameraRetentionEstimates() {
			var bytesAvailable,
				bytesFromOverrides,
				bytesPerDay,
				bytesPerMinute,
				daysEstimated;

			// calculate bytes used up by overrides
			bytesFromOverrides = $scope.configuredCameras.reduce(function(val, camera) {
				var cameraMeta = $scope.configuredCamerasMeta[camera.id];

				if (cameraMeta.override === true) {
					return val + cameraMeta.bytesEstimatedPerDay * ((angular.isUndefined(cameraMeta.daysDesired) === true) ? 0 : cameraMeta.daysDesired);
				}
				else {
					return val;
				}
			}, 0);

			// calculate true bytes available
			bytesAvailable = $scope.diskUtes.availableUpToThreshold - bytesFromOverrides;

			// calculate total bytes per day for non-overridden cameras
			bytesPerDay = $scope.configuredCameras.reduce(function(val, camera) {
				if ($scope.configuredCamerasMeta[camera.id].override === false) {
					return val + $scope.configuredCamerasMeta[camera.id].bytesEstimatedPerDay;
				}
				else {
					return val;
				}
			}, 0);

			// and calculate bytes per minute so we can fill in the gaps
			bytesPerMinute = bytesPerDay / 24 / 60;

			// using our calculated bytesAvailable, determine how many days all automatic cameras will get
			daysEstimated = 0;

			if (bytesPerDay > 0 && bytesAvailable > 0) {
				// use bytesPerDay first because it should be a more accurate value (is this actually true?? what if bytesPerMinute were just based on the per day value??)
				daysEstimated = Math.floor(bytesAvailable / bytesPerDay);

				// adjust bytesAvailable so we can fill in the gaps with minutes
				bytesAvailable -= bytesPerDay * daysEstimated;

				// fill in the gaps with hours
				daysEstimated += 1 / 24 / 60 * Math.floor(bytesAvailable / Math.floor(bytesPerMinute));
			}

			// set bytesRecordedPercentage
			$scope.configuredCameras.forEach(function(camera) {
				var cameraMeta = $scope.configuredCamerasMeta[camera.id],
					days = ((cameraMeta.override === false) ? daysEstimated : ((angular.isUndefined(cameraMeta.daysDesired) === true) ? 0 : cameraMeta.daysDesired));

				cameraMeta.daysEstimated = days;
				cameraMeta.bytesEstimated = cameraMeta.bytesEstimatedPerDay * days;
				cameraMeta.bytesEstimatedPercentage = cameraMeta.bytesEstimated / $scope.diskUtes.capacity * 100;
			});

			$scope.percentageEstimated = $scope.configuredCameras.reduce(function(val, camera) {
				return val + $scope.configuredCamerasMeta[camera.id].bytesEstimatedPercentage;
			}, $scope.diskUtes.unavailablePercentage);
		}

		$scope.toggleRetentionOverride = function(cameraMeta) {
			cameraMeta.override = !cameraMeta.override;

			if (cameraMeta.override === true) {
				cameraMeta.daysDesired = parseInt(cameraMeta.daysEstimated, 10);
				FocusOnService('override-' + cameraMeta.id);
			}

			$scope.retentionPolicyForm.$setDirty();
		};

		$scope.revertChanges = function() {
			$scope.configuredCamerasMeta = angular.copy($scope.configuredCamerasMetaOriginal);
			$scope.retentionPolicyForm.$setPristine();
		};

		$scope.saveConfiguration = function() {
			var savePromises = [];

			$scope.processingSave = true;

			$scope.configuredCameras.forEach(function(configuredCamera) {
				var cameraMeta = $scope.configuredCamerasMeta[configuredCamera.id],
					camera = {
						retention: ((cameraMeta.override === true) ? cameraMeta.daysDesired * 24 * 60 * 60 * 1000 : 0)
					};

				savePromises.push($http.patch(BaseServiceUrl + '/service/cameras/' + configuredCamera.id, camera).then(function() {
					configuredCamera.retention = camera.retention;
				}));
			});

			$q.all(savePromises)
				.then(function() {
					$scope.processingSave = false;
					$scope.processingSaveFailed = false;
					$scope.processingSaveFailedMessage = '';

					$scope.retentionPolicyForm.$setPristine();

					// update the original so we can revert from the right point
					$scope.configuredCamerasMetaOriginal = angular.copy($scope.configuredCamerasMeta);
				})
				.catch(function(badResponse) {
					$scope.processingSave = false;

					$scope.processingSaveFailed = true;
					$scope.processingSaveFailedMessage = 'An unexpected error occurred when saving retention';

					if (Array.isArray(badResponse) && badResponse.length > 0 && badResponse[0].data) {
						$scope.processingSaveFailedMessage += ': ' + badResponse[0].data.reason;
					}
					else {
						$scope.processingSaveFailedMessage += '.';
					}
				});
		};

		// init
		$scope.$on('$destroy', function() {
			$scope.destroyed = true;
		});

		$scope.$watch('configuredCamerasMeta', function() {
			calculateCameraRetentionEstimates();
		}, true);

		$scope.processing = true;
		$scope.processingFailed = false;
		$scope.processingFailedMessage = '';
		$scope.processingSave = false;
		$scope.processingSaveFailed = false;
		$scope.processingSaveFailedMessage = '';
		$scope.configuredCameras = [];
		$scope.configuredCamerasMeta = {};
		$scope.configuredCamerasMetaOriginal = {};
		$scope.diskUtes = {};
		$scope.currentInstance = OrchidDiscoveryService.currentInstance;

		// via http://jsfiddle.net/langdonx/ty25uoad/
		$scope.colors = ['#32ff32', '#25bf25', '#167216', '#3232ff', '#2525bf', '#161672', '#e6e600', '#acac00', '#676700', '#00e6e6', '#00acac', '#006767', '#e600e6', '#ac00ac', '#670067', '#ffa500', '#bf7b00', '#724a00', '#ff00a5', '#bf007b', '#72004a', '#00ffa5', '#00bf7b', '#00724a', '#a5ff00', '#7bbf00', '#4a7200', '#00a5ff', '#007bbf', '#004a72', '#a500ff', '#7b00bf', '#4a0072', '#ff3232', '#bf2525', '#721616'];

		// load all required data and use $q.all to wait for it
		// 1. configuredCameras ($q)
		// 2. camera retention ($http)
		// 3. disk utes ($http)
		promises.push(fetchConfiguredCameras());
		promises.push(fetchDiskUtes());
		promises.push(fetchRetention());

		$q.all(promises)
			.then(function(promiseResponses) {
				if ($scope.destroyed !== true) {
					// set cameras, and sort them here, once, because it's more efficient than sorting it 3 times later and gets us the color choices we want
					$scope.configuredCameras = $filter('orderBy')(promiseResponses[0], 'name');

					// set disk utes
					initializeDiskUtes(promiseResponses[1].data);

					// initialize camera metadata collection
					initializeCameraMeta();

					// set camera retention data
					initializeCameraRetention(promiseResponses[2].data, promiseResponses[1].data);

					// estimate days of retention
					calculateCameraRetentionEstimates();

					$scope.configuredCamerasMetaOriginal = angular.copy($scope.configuredCamerasMeta);
					$scope.processing = false;
				}
			})
			.catch(function(response) {
				if ($scope.destroyed !== true) {
					$scope.processing = false;
					$scope.processingFailed = true;

					if (typeof response === 'string') {
						$scope.processingFailedMessage = response;
					}
					else if (response && response.data && typeof response.data.reason === 'string') {
						$scope.processingFailedMessage = response.data.reason;
					}
					else {
						$scope.processingFailedMessage = 'Unknown error occurred.';
					}

					$log.error('init error', response);
				}
			});
	}
}());
