(function() {
	'use strict';

	angular.module('ipconfigure.orchid').controller('AdminCtrl', ['$scope', '$http', '$log', '$timeout', 'AuthService', 'BaseServiceUrl', 'CameraService', AdminCtrl]);

	function AdminCtrl($scope, $http, $log, $timeout, AuthService, BaseServiceUrl, CameraService) {
		var discoverTimer;

		$scope.configuredCamerasLoading = true;
		$scope.discoveredCamerasLoading = true;
		$scope.configuredCameras = [];
		$scope.discoveredCameras = [];
		$scope.included = 'admin-version';
		$scope.selectedDiscoveredCamera = undefined;
		$scope.selectedConfiguredCamera = undefined;

		$scope.$on('change-admin-include', function(event, data) {
			$scope.included = data;
		});

		$scope.goCameraConfiguration = function() {
			$scope.included = 'admin-camera-configuration';
		};

		$scope.goCameraAdd = function() {
			$scope.included = 'admin-camera-add';
		};

		$scope.goCameraDiscover = function(camera) {
			$scope.included = 'admin-camera-discover';

			$scope.selectedDiscoveredCamera = camera;
		};

		$scope.goCameraEdit = function(camera) {
			$scope.included = 'admin-camera-edit';

			$scope.selectedConfiguredCamera = camera;
			$scope.selectedConfiguredCameraModel = (camera.driver === 'ONVIF') ? camera.configuration.Manufacturer + ' ' + camera.configuration.Model.replace(camera.configuration.Manufacturer + ' ', '') : 'Unknown, Generic RTSP';
		};

		$scope.goRecordingSchedules = function() {
			$scope.included = 'admin-schedule';
		};

		$scope.goActivation = function() {
			$scope.included = 'admin-activation';
		};

		$scope.goSystemReport = function() {
			$scope.included = 'admin-system-report';
		};

		$scope.goUserManagement = function() {
			$scope.included = 'admin-users';
		};

		$scope.goSystemConfiguration = function() {
			$scope.included = 'admin-system';
		};

		$scope.goVersion = function() {
			$scope.included = 'admin-version';
		};

		$scope.$watch('adminMode', function(newValue) {
			if (newValue === true) {
				(function discoverLoop() {
					// refresh discoverable cameras
					if ($scope.adminMode === true) {
						_fetchDiscoveredCameras(function() {
							discoverTimer = $timeout(discoverLoop, 10000);
						});
					}
				}());
			}
		});

		$scope.$watch('cameraServiceState.currentState', function(currentState) {
			if (currentState !== 'loading') {
				$scope.configuredCamerasLoading = false;
			}
		});

		$scope.$on('$destroy', function() {
			$timeout.cancel(discoverTimer);
		});

		function _fetchDiscoveredCameras(callback) {
			$http.get(BaseServiceUrl + '/service/discoverable/cameras')
				.success(function(response) {
					$scope.discoveredCameras = [];

					if (response.cameras) {
						response.cameras.forEach(function(camera) {
							// TODO determine if camera is already registered
							camera.registered = $scope.configuredCameras.some(function(configuredCamera) {
								if (configuredCamera.configuration.IPAddress === camera.ip) {
									return true;
								}
							});

							$scope.discoveredCameras.push(camera);
						});
					}

					$scope.discoveredCamerasLoading = false;

					callback();
				})
				.error(function(data, status, headers, config) {
					if (status !== 0) {
						$log.error(data, status, headers, config);

						callback();
					}
				});
		}

		function _init() {
			var user = AuthService.getAuthenticatedUser();

			if (user.role !== 'Administrator' && user.role !== 'Manager') {
				// redirect in case they somehow accessed this controller (can happen via configuration.html)
				location.href = '/';
			}
			else {
				$scope.configuredCameras = CameraService.cameras;
				$scope.cameraServiceState = CameraService.state;

				// TODO put in a directive
				$timeout(function() {
					var visibleSections;
					if (document.body.clientWidth < 992) {
						document.querySelector('.camera-list-container').style.height = '120px';
					}
					else {
						visibleSections = document.querySelectorAll('.admin-nav > li:not(.ng-hide)').length;
						document.querySelector('.camera-list-container').style.height = 'calc(100vh - 69px - ' + (visibleSections * 49) + 'px + 5px)';
					}
				});
			}
		}

		_init();
	}
}());
