(function() {
	'use strict';

	angular.module('ipconfigure.orchid').controller('AdminUsersCtrl', ['$scope', '$http', '$log', 'BaseServiceUrl', 'FocusOnService', AdminUsersCtrl]);

	function AdminUsersCtrl($scope, $http, $log, BaseServiceUrl, FocusOnService) {
		$scope.loading = true;
		$scope.processing = false;
		$scope.processingDelete = {};
		$scope.failure = '';
		$scope.selectedUser = undefined;
		$scope.users = [];

		function _init() {
			$scope.loading = true;

			$http.get(BaseServiceUrl + '/service/users')
				.success(function(response) {
					$scope.loading = false;

					$scope.users = response.users;
				})
				.error(function(data, status, headers, config) {
					$log.error(data, status, headers, config);

					$scope.loading = false;

					if (status !== 0) {
						window.alert('user request failed');
					}
				});
		}

		function fixPassword(user) {
			// because of how the web server was implemented, if password is empty string, it thinks you're trying to
			// change the password to empty string (instead of simply ignoring it).  this cleans up the data before
			// it goes to the web server (called by both addUser and updateUser).
			if (!user.password) {
				delete user.password;
			}
		}

		$scope.addUser = function(user) {
			if ((user.password || user.passwordVerify) && user.password !== user.passwordVerify) {
				$scope.failure = 'Passwords do not match';
			}
			else {
				$scope.processing = true;

				fixPassword(user);

				$http.post(BaseServiceUrl + '/service/users', user)
					.success(function(response) {
						$scope.processing = false;

						$scope.users.push(response);
						$scope.selectedUser = undefined;
					})
					.error(function(data, status, headers, config) {
						$log.error(data, status, headers, config);

						$scope.processing = false;

						if (data && (data.code || data.reason)) {
							$scope.failure = data.code + ': ' + data.reason;
						}
						else {
							$scope.failure = 'An unexpected error occurred.';
						}
					});
			}
		};

		$scope.updateUser = function(user, shadowCommitFn) {
			if ((user.password || user.passwordVerify) && user.password !== user.passwordVerify) {
				$scope.failure = 'Passwords do not match';
			}
			else {
				$scope.processing = true;

				fixPassword(user);

				$http.patch(BaseServiceUrl + '/service/users/' + user.id, user)
					.success(function(response) {
						$scope.processing = false;

						// not entirely necessary, since we're taking the server response as gospel
						shadowCommitFn();

						// find and update user
						$scope.users.some(function(someUser, index) {
							if (someUser.id === user.id) {
								$scope.users[index] = response;
								return true;
							}
						});

						$scope.selectedUser = undefined;
					})
					.error(function(data, status, headers, config) {
						$log.error(data, status, headers, config);

						$scope.processing = false;

						if (data && (data.code || data.reason)) {
							$scope.failure = data.code + ': ' + data.reason;
						}
						else {
							$scope.failure = 'An unexpected error occurred.';
						}
					});
			}
		};

		$scope.deleteUser = function(user) {
			if (window.confirm('Are you sure?') === true) {
				$scope.processingDelete[user.id] = true;

				$http.delete(BaseServiceUrl + '/service/users/' + user.id)
					.success(function() {
						$scope.processingDelete[user.id] = false;

						// find user in array and splice it
						$scope.users.some(function(scopeUser, i) {
							if (scopeUser.id === user.id) {
								$scope.users.splice(i, 1);
								return true;
							}
						});

						if ($scope.selectedUser && $scope.selectedUser.id === user.id) {
							$scope.selectedUser = undefined;
						}
					})
					.error(function(data, status, headers, config) {
						$log.error(data, status, headers, config);

						$scope.processingDelete[user.id] = false;

						if (status !== 0) {
							window.alert('delete user request failed');
						}
					});
			}
		};

		$scope.makeNewUser = function() {
			return {
				role: 'Viewer'
			};
		};

		$scope.selectUser = function(user) {
			if ($scope.processing === false) {
				$scope.failure = '';
				$scope.selectedUser = user;

				$scope.userForm.$setPristine();

				FocusOnService('initial-field');
			}
		};

		_init();
	}
}());
