(function() {
	'use strict';

	angular.module('ipconfigure.orchid')
		.controller('ExportCtrl', ['$scope', '$filter', '$modalInstance', '$http', 'AuthService', 'PlayerComponentRegistrar', 'TimeDeltaService', 'fineTime', ExportCtrl]);

	function ExportCtrl($scope, $filter, $modalInstance, $http, AuthService, PlayerComponentRegistrar, TimeDeltaService, fineTime) {
		var exportLength = fineTime[1] - fineTime[0],
			uniqueStreams = {};

		$scope.user = AuthService.getAuthenticatedUser();
		$scope.length = $filter('timeDuration')(exportLength, true);

		if (exportLength > 1000 * 60 * 60) {
			$scope.invalidRange = true;
		}
		else {
			$scope.streams = [];

			PlayerComponentRegistrar.getComponents().some(function(components) {
				if (components.playerScope.fullScreen === true) {
					// if we encounter a player that is full screen, include only it
					// on the export screen and nothing else
					$scope.streams = [angular.copy(components.playerScope.stream)];
					uniqueStreams = {};
					uniqueStreams[components.playerScope.stream.id] = true;
					return true;
				}
				else {
					if (uniqueStreams[components.playerScope.stream.id] !== true) {
						$scope.streams.push(angular.copy(components.playerScope.stream));
					}

					uniqueStreams[components.playerScope.stream.id] = true;
				}
			});
		}


		$scope.exportStream = function(stream, format) {
			var startTime = fineTime[0] - TimeDeltaService.getDifference(),
			endTime = fineTime[1] - TimeDeltaService.getDifference();
			$http.get(stream.cameraExpanded.instance.baseUrl + '/service/archives?take=1&streamId=' + stream.id + '&start=' + (fineTime[0] - TimeDeltaService.getDifference())).success(function(response) {
				if(response.archives !== null && (response.archives.length > 0) && (response.archives[0].end < endTime || response.archives[0].start < endTime)) {
					window.open(stream.cameraExpanded.instance.baseUrl + '/service/streams/' + stream.id + '/export?start=' + startTime + '&stop=' + endTime + '&format=' + format);

					stream.exported = true;
				}
				else {
					alert("There are no archives in the specified region.");
				}
			}).error(function() {
				alert("An error has occurred while attempting to export video.");
			});
		}

		$scope.cancel = function() {
			$modalInstance.dismiss('cancel');
		};
	}
}());
