(function() {
	'use strict';

	angular.module('ipconfigure.orchid')
		.controller('SoulCtrl', ['$scope', '$document', '$filter', '$log', '$rootScope', '$timeout', '$controller', '$compile', '$uibModal', 'AuthService', 'MotionDetectionService', 'PlayerComponentRegistrar', 'StageService', 'TimeDeltaService', SoulCtrl])
		.controller('ChildCtrl', ['$scope', '$rootScope', '$log', ChildCtrl]);

	function SoulCtrl($scope, $document, $filter, $log, $rootScope, $timeout, $controller, $compile, $uibModal, AuthService, MotionDetectionService, PlayerComponentRegistrar, StageService, TimeDeltaService) {
		var _liveTolerance = 5000,
			_id = 0,
			_masterTimeline,
			_monitorPlayerLastNow,
			_monitorPlayerTimer,
			_monitorMotionEventsTimer,
			_playbackFromTime,
			_fullScreen = false,
			_background = {
				time: Date.now(),
				live: false,
				paused: false,
				fullscreenLinked: false,
				masterMode: 'live'
			};

		$scope.masterMode = 'live';

		$scope.scales = [
			'1 day',
			'12 hr',
			'6 hr',
			'3 hr',
			'1 hr',
			'30 min',
			'15 min',
			'5 min',
			'1 min'
		];
		$scope.scaleAdjusting = false;
		$scope.scaleIndex = $scope.scales.length - 1;

		$scope.paused = false;
		$scope.pausedAt = 0;
		$scope.pauseOffset = 0;

		$scope.motionEventProcessing = false;

		$scope.datePicker = false;

		$scope.players = [];
		$scope.streams = [];

		$scope.dt = ipConfigure.core.floorDateToMinutes(new Date(), 1);
		$scope.maxDate = new Date();

		$scope.role = AuthService.getAuthenticatedUser().role;

		$scope.previousButtonShake = 0;
		$scope.nextButtonShake = 0;

		$scope.addPlayer = function(stream, container) {
			var playFromTime,
				html,
				locals = { $scope: angular.element(container.element[0][0]).scope().$new() };

			if (_playbackFromTime) {
				playFromTime = _masterTimeline.getTime();
			}
			else {
				if ($scope.paused && $scope.pausedAt) {
					playFromTime = $scope.pausedAt;
				}
				else {
					if (-($scope.pauseOffset) > _liveTolerance) {
						playFromTime = Date.now() + $scope.pauseOffset;
					}
				}
			}

			$scope.players.push({
				id: ++_id,
				playFromTime: playFromTime,
				cell: container
			});

			html = '<player ' +
				' player-id="playerId"' +
				' stream="stream"' +
				' instance="instance"' +
				((playFromTime) ? 'play-from-time="' + playFromTime + '"' : '') +
				' auto-pause="paused"' +
				' onremove="removePlayer()"' +
				' onfullscreen="toggleFullScreen(playerId, fullScreen)">' +
				'</player>';

			locals.$scope.playerId = _id;
			locals.$scope.stream = stream;
			locals.$scope.instance = stream.cameraExpanded.instance;
			locals.$scope.paused = $scope.paused;

			$controller('ChildCtrl', locals);

			// keep track of which stream is in which cell for stage persistence
			container.streamId = stream.id;

			// append the new html to the body
			container.element.node().innerHTML = '';

			angular.element(container.element.node()).append(html);

			// compile that mama
			$compile(container.element.node().firstChild)(locals.$scope);

			// save
			StageService.persistStageConfiguration();
		};

		$scope.getMotionEvent = function(direction, fromBeginning) {
			var buttonShake,
			fromTime,
			fetchMotionEvent,
			directionIndex = (direction === 'next') ? 0 : 1;

			if ($scope.role !== 'Live Viewer') {
				fromTime = (!!fromBeginning && $scope.masterMode === 'selection') ? _masterTimeline.getFineTime()[directionIndex] : fromTime = _masterTimeline.getTime();
				$scope.motionEventProcessing = true;
				MotionDetectionService.fetchMotionEvent(direction, fromTime, function(result) {
					$scope.motionEventProcessing = false;

					if (result === false || ($scope.masterMode === 'selection' && !_masterTimeline.hasTime(result))) {
						if (!fromBeginning && $scope.masterMode === 'selection') {
							// if we're in selection mode, loop around to the other side and try again
							$scope.getMotionEvent(direction, true);
						}
						else {
						// shake dem buttons
							(direction === 'next') ? $scope.nextButtonShake++ : $scope.previousButtonShake++;
						}
					}
					else {
						_setupTimelineAndPlayersAtTime(result, 'via-motion-event');
					}
				});
			}
		};

		$scope.playPause = function() {
			var playerTime,
				playerComponents;

			if ($scope.role !== 'Live Viewer') {
				playerTime = _playbackFromTime ? _playbackFromTime : Date.now();
				playerComponents = PlayerComponentRegistrar.getComponents();

				if ($scope.paused) {
					$scope.pauseOffset -= playerTime - $scope.pausedAt + $scope.pauseOffset;
				}

				$scope.paused = !$scope.paused;
				$scope.pausedAt = playerTime + $scope.pauseOffset;

				if (!$scope.paused) {
					// if the user un-paused and we're within the live threshold, just jump to live
					if (_isLive($scope.pausedAt)) {
						$scope.pausedAt = 0;
						$scope.pauseOffset = 0;
					}
				}

				$log.debug('pausedAt', $scope.pausedAt, Date($scope.pausedAt));
				$log.debug('pauseOffset', $scope.pauseOffset);

				playerComponents.forEach(function(e) {
					if ((e.video.linked !== false) && (!_fullScreen || e.playerScope.fullScreen)) {
						if ($scope.paused === true) {
							e.video.stop();
						}
						else {
							if ($scope.pausedAt === 0) {
								e.video.playLive();
							}
							else {
								e.video.playFromTime($scope.pausedAt, false);
							}
						}
					}
				});
			}
		};

		$scope.jumpLeft = function() {
			var coarseRange = _masterTimeline.getCoarseTime(),
				diff;

			if (typeof coarseRange[0] === 'undefined') {
				$log.debug('requested coarse time when not in coarse mode');
			}
			else {
				diff = coarseRange[0] - coarseRange[1];

				$log.debug('coarse', new Date(coarseRange[0]), new Date(coarseRange[1]));
				$log.debug('go', new Date(coarseRange[0] + diff), new Date(coarseRange[1] + diff));

				if ($scope.paused) {
					$scope.playPause();
				}

				_masterTimeline.setCoarseTime(coarseRange[0] + diff, coarseRange[1] + diff);
			}
		};

		$scope.jumpRight = function() {
			var coarseRange = _masterTimeline.getCoarseTime(),
				diff;

			if (typeof coarseRange[0] === 'undefined') {
				$log.debug('requested coarse time when not in coarse mode');
			}
			else {
				diff = coarseRange[1] - coarseRange[0];

				$log.debug('coarse', new Date(coarseRange[0]), new Date(coarseRange[1]));
				$log.debug('go', new Date(coarseRange[0] + diff), new Date(coarseRange[1] + diff));

				if ($scope.paused) {
					$scope.playPause();
				}

				_masterTimeline.setCoarseTime(coarseRange[0] + diff, coarseRange[1] + diff);
			}
		};

		$scope.scaleLeft = function() {
			if ($scope.scaleAdjusting === false && $scope.scaleIndex < $scope.scales.length - 1) {
				$scope.scaleAdjusting = true;

				_masterTimeline.adjustScale(1, function() {
					$scope.scaleAdjusting = false;
				});
			}
		};

		$scope.scaleRight = function() {
			if ($scope.scaleAdjusting === false && $scope.scaleIndex > 0) {
				$scope.scaleAdjusting = true;

				_masterTimeline.adjustScale(-1, function() {
					$scope.scaleAdjusting = false;
				});
			}
		};

		$scope.goDate = function() {
			_setupTimelineAndPlayersAtTime($scope.dt.getTime(), 'date-picker');
			$scope.masterMode = 'live';
			$scope.datePicker = false;
			_masterTimeline.trackLive();
		};

		$scope.toggleDatePicker = function() {
			$scope.dt = ipConfigure.core.floorDateToMinutes(new Date(_masterTimeline.getFineTime()[0]), 1);
			$scope.datePicker = !$scope.datePicker;
		};

		$scope.export = function() {
			var fineTime;

			if ($scope.masterMode === 'selection' && $scope.role !== 'Viewer' && $scope.role !== 'Live Viewer') {
				fineTime = _masterTimeline.getFineTime();
				$uibModal.open({
					templateUrl: '/views/export.html',
					controller: 'ExportCtrl',
					backdrop: 'static',
					resolve: {
						fineTime: function() {
							return fineTime;
						}
					}
				});
			}
		};

		$scope.expandTimeline = function() {
			if ($scope.role !== 'Live Viewer') {
				if ($scope.masterMode !== 'live') {
					_masterTimeline.trackLive();
					$scope.masterMode = 'live';
					$scope.datePicker = false;
				}
				else {
					_masterTimeline.trackPlayback();
					$scope.masterMode = 'selection';
				}
			}
		};

		$scope.$on('$destroy', function() {
			//if (typeof _cancelOnUserIdle === 'function') {
			//	_cancelOnUserIdle();
			//}

			_masterTimeline.destroy();

			$timeout.cancel(_monitorPlayerTimer);
			$timeout.cancel(_monitorMotionEventsTimer);
		});

		$scope.$on('chill-mode', function(event, active) {
			$scope.datePicker = false;

			// tell the _masterTimeline about chill-mode
			_masterTimeline.anchorMotionEventsToTop(active);
		});

		$scope.$on('relink', function(event, playerId) {
			var component = PlayerComponentRegistrar.getComponentsByPlayerId(playerId);
			//
			//				$scope.players.some(function (e) {
			//					if (e.id === playerId) {
			//						e.linked = false;
			//						return true;
			//					}
			//					else {
			//						return false;
			//					}
			//				});

			// when a video is relinked to the master timeline, set the video appropriately
			if (_playbackFromTime) {
				component.video.playFromTime(_masterTimeline.getTime());
			}
			else {
				// only reset the video if
				// it's not already streaming live
				if (component.video.isLive !== true) {
					component.video.playLive();
				}
			}
		});

		$scope.$on('removePlayer', function(event, playerId, maintainContainer) {
			PlayerComponentRegistrar.unregisterComponent(playerId);

			$scope.players.some(function(e, i) {
				if (e.id === playerId) {
					// properly $destroy the player...
					angular.element(e.cell.element.node().firstElementChild).scope().$destroy();

					// and remove contents from the cell
					e.cell.element.node().innerHTML = '';

					// clean up any modifications we made to the stage's cell
					delete e.streamId;

					if (maintainContainer !== true) {
						StageService.releaseCell(e.cell);
						StageService.persistStageConfiguration();
					}

					$scope.players.splice(i, 1);
					return true;
				}
			});
		});

		$scope.$on('fullScreen', function(event, playerId) {
			var component = PlayerComponentRegistrar.getComponentsByPlayerId(playerId),
				maximizing = component.playerScope.fullScreen,
				time;

			if (maximizing) {
				// We are going to re-link a fullscreen unlinked player later so that it can use the master timeline while
				// in full screen mode.  Capture its state so we can re-un-link it later.
				_background.fullscreenLinked = component.playerScope.linked;
			}
			_fullScreen = maximizing;

			if (_background.fullscreenLinked === false) {
				component.playerScope.toggleLink();
				if (maximizing) {

					// record the state of the master timeline and update it with the fullscreening player's state
					_background.live = (_isLive(_masterTimeline.getTime()));
					_background.time = _masterTimeline.getTime();
					_background.paused = $scope.paused;
					_background.masterMode = $scope.masterMode;

					_masterTimeline.setFineTime(component.timeline.getTime());
					$scope.masterMode = 'live'; // unlinked players don't have a selection mode
					$scope.paused = false; // unlinked players cannot be paused
					_setupTimelineAndPlayersAtTime(component.timeline.getTime(), 'via-fullscreen');
				}
				else {
					// restore the master timeline state
					$scope.paused = _background.paused;
					component.timeline.setTime(_masterTimeline.getTime());

					time = (_background.live || _isLive(_background.time)) ? Date.now() : _background.time;
					_setupTimelineAndPlayersAtTime(time, 'via-fullscreen');
					_masterTimeline.setFineTime(time);
					if (_background.masterMode === 'selection') {
						_masterTimeline.trackPlayback();
					}
					$scope.masterMode = _background.masterMode;
				}
			}
			else {
				_background.live = true;
				_background.paused = $scope.paused;
			}

			PlayerComponentRegistrar.getComponents().forEach(function(e) {
				var timeline;
				if (!(e.playerScope.playerId === playerId && maximizing)) {
					if (maximizing) {
						// pause all players except for the fullscreen player
						if (e.playerScope.linked === false && (_isLive(e.timeline.getTime()))) {
							e.playerScope.backgroundLive = true;
						}
						e.video.stop();
					}
					else { // restoring player
						if (!e.playerScope.linked || !$scope.paused) {
							timeline = e.playerScope.linked ? _masterTimeline : e.timeline;

							// play each player at either live or its playback time
							if (e.playerScope.backgroundLive === true || (component.playerScope.linked === false && e.playerScope.linked && _background.live) || _isLive(timeline.getTime())) {
								e.playerScope.backgroundLive = false;
								e.video.playLive();
							}
							else {
								e.video.playFromTime(timeline.getTime());
							}
						}
					}
				}
			});

			// tell the stage to make this cell full screen
			$scope.players.some(function(e) {
				if (e.id === playerId) {
					StageService.toggleFullScreen(e.cell);
				}
			});
			// force hide the camera column
			$scope.$broadcast('bag-force-close');
		});

		$scope.$watch('masterMode', function(newValue, oldValue) {
			if (newValue !== oldValue) {
				// allow for browser to render the changes
				$timeout(function() {
					$rootScope.$broadcast('resize', 'master-mode');
				}, 20);
			}
		});

		// create master-timeline
		_masterTimeline = new ipConfigure.components.timeline('master-timeline', {
			debug: false,
			interactive: ($scope.role !== 'Live Viewer'),
			onTimeChange: function(time, source) {
				_setupTimelineAndPlayersAtTime(time, source);

				// since we've elected to not invokeApply with our $timeout, this is necessary
				$scope.$apply();
			},
			onCoarseTimeChange: function(startTime, stopTime) {
				var startDate = new Date(startTime),
					stopDate = new Date(stopTime);

				$log.debug('onCoarseTimeChange', startTime, stopTime);

				// if startTime and stopTime have different months
				if (startDate.getMonth() !== stopDate.getMonth()) {
					$scope.dateRange = $filter('date')(startTime, 'MMM') + ' - ' + $filter('date')(stopTime, 'MMM, yyyy');
				}
				// if startTime and stopTime have different days
				else if (startDate.getDate() !== stopDate.getDate()) {
					$scope.dateRange = $filter('date')(startTime, 'MMM d') + ' - ' + $filter('date')(stopTime, 'd, yyyy');
				}
				else {
					$scope.dateRange = $filter('date')(startDate, 'MMMM d, yyyy');
				}
			},
			onEndReached: function(timeWindowExtent) {
				var i,
					playerComponents = PlayerComponentRegistrar.getComponents();

				_playbackFromTime = timeWindowExtent[0];

				// update linked players
				for (i = 0; i < playerComponents.length; i++) {
					if (playerComponents[i].video.linked !== false) {
						playerComponents[i].video.playFromTime(timeWindowExtent[0]);
					}
				}
			},
			onScaleAdjust: function(configurationIndex) {
				$scope.scaleIndex = configurationIndex;
			},
			onRip: function(time, metadata) {
				var playerComponents = PlayerComponentRegistrar.getComponents();

				playerComponents.forEach(function(component) {
					if (component.video.linked !== false) {
						component.playerScope.rip(time, metadata);
					}
				});
			},
			onFineDblClick: function() {}
		});

		_masterTimeline.trackLive();

		$rootScope.$on('resize', function() {
			$timeout(function() {
				_masterTimeline.resize();
			}, 50);
		});

		// kick off the timer that will keep all timelines ticking
		(function _monitorPlayer() {
			var playerComponents = PlayerComponentRegistrar.getComponents(),
				now = Date.now(),
				i,
				playTime,
				currentTime;

			// update timeline for non-linked players
			for (i = 0; i < playerComponents.length; i++) {
				// only fire if video's video is not linked to the master timeline and the overlay/timeline is visible
				if (playerComponents[i].video.linked === false) {
					if (playerComponents[i].video.isLive === true) {
						playerComponents[i].timeline.setTime(now);
					}
					else {
						playTime = playerComponents[i].video.getCurrentTime();

						if (playTime > 0) {
							playerComponents[i].timeline.setTime(playTime);
						}
					}
				}
			}

			// update master timeline
			if ($scope.paused) {
				currentTime = $scope.pausedAt;
			}
			else {
				if (_playbackFromTime) {
					_playbackFromTime += (now - _monitorPlayerLastNow);

					currentTime = _playbackFromTime;
				}
				else {
					currentTime = now + $scope.pauseOffset;
				}
			}

			PlayerComponentRegistrar.setTime(currentTime);
			_masterTimeline.setTime(currentTime);
			_monitorPlayerLastNow = now;

			// 25 fps
			_monitorPlayerTimer = $timeout(_monitorPlayer, 1000 / 25, false);
		}());

		// kick off the timer that will fetch motion event data for all timelines
		(function _monitorMotionEvents() {
			MotionDetectionService.setMotionDetectionAreas(_masterTimeline, function() {
				_monitorMotionEventsTimer = $timeout(_monitorMotionEvents, 1500, false);
			});
		}());

		// now that we're authenticated, we need to fetch the server time
		TimeDeltaService.fetchDifferenceInfo().then(function(differenceInfo) {
			// unfortunately, we need to send this data to AppCtrl's scope... so this is awful
			$scope.$parent.$parent.clientServerTimeDifference = differenceInfo.diff;
		});

		$document.bind('keydown', function(event) {
			if (event.keyCode === 27) {
				$scope.datePicker = false;
			}
		});

		function _isLive(time) {
			return Date.now() - time < _liveTolerance;
		};

		function _setupTimelineAndPlayersAtTime(time, source) {
			var i,
				diff = Date.now() - time,
				playerComponents = PlayerComponentRegistrar.getComponents();

			if (_fullScreen === false && source !== 'via-fullscreen') {
				$scope.paused = false;
				$scope.pausedAt = 0;
				$scope.pauseOffset = 0;
			}

			if (diff < _liveTolerance) {
				_playbackFromTime = undefined;

				$scope.datePicker = false;

				_masterTimeline.setFineTime(time);
				if(source !== 'via-fullscreen') {
					$scope.masterMode = 'live';
					_masterTimeline.trackLive();
				}

				// update linked players
				for (i = 0; i < playerComponents.length; i++) {
					if ((_fullScreen === false || playerComponents[i].playerScope.fullScreen === true) && (playerComponents[i].video.linked !== false)) {
						playerComponents[i].playerScope.rip(false, { source: source });
						playerComponents[i].video.playLive();
					}
				}
			}
			else {
				_playbackFromTime = time;

				// animate to specific time
				if (source === 'via-fine-click' || source === 'via-motion-event' || source === 'date-picker') {
					_masterTimeline.setFineTime(time);
				}

				// enter into selection mode
				if (source !== 'via-fine' && source !== 'via-fine-click' && source !== 'via-motion-event' && source !== 'date-picker' && source !== 'via-fullscreen') {
					_masterTimeline.trackPlayback();
					$scope.masterMode = 'selection';
				}

				// update linked players
				for (i = 0; i < playerComponents.length; i++) {
					if ((_fullScreen === false || playerComponents[i].playerScope.fullScreen === true) && playerComponents[i].video.linked !== false) {
						playerComponents[i].playerScope.rip(false, { source: source });
						playerComponents[i].video.playFromTime(time);
					}
				}
			}
		}
	}

	function ChildCtrl($scope, $rootScope, $log) {
		$scope.removePlayer = function() {
			$rootScope.$broadcast('removePlayer', $scope.playerId);
		};

		$scope.toggleFullScreen = function(playerId, fullScreen) {
			$log.debug('playerId: ' + playerId + '; fullScreen: ' + fullScreen);
		};
	}
}());
