(function() {
	'use strict';

	angular.module('ipconfigure.orchid.directive').directive('chillMode', ['$log', '$rootScope', '$timeout', 'IdleService', 'PlayerComponentRegistrar', chillMode]);

	function chillMode($log, $rootScope, $timeout, IdleService, PlayerComponentRegistrar) {
		return {
			restrict: 'A',
			link: function(scope) {
				var _chillModeActivateAfter = 5, //seconds
					_mode = 'wile',
					_elements = {
						header: document.getElementsByTagName('header')[0],
						bag: document.getElementById('bag'),
						footer: document.getElementsByTagName('footer')[0],
						stage: document.getElementById('stage')
					};

				function _isFullScreen() {
					return (window.fullScreen ||
						(window.innerWidth === screen.width && window.innerHeight === screen.height) ||
						window.navigator.standalone);
				}

				function _chillOut() {
					$log.debug('chillOut');

					_mode = 'chill';

					_elements.bag.style.height = '100%';
					_elements.bag.style.opacity = 0;
					_elements.footer.style.bottom = '-' + (parseInt(_elements.footer.style.height, 10) - 24) + 'px';
					Array.prototype.slice.call(_elements.header.querySelectorAll('i.fa')).forEach(function(i) {
						i.style.visibility = 'hidden';
					});

					// adjust stage
					_elements.stage.style.height = 'calc(100% - 36px)';

					$rootScope.$broadcast('chill-mode', true);

					$timeout(function() {
						$rootScope.$broadcast('bag-force-close');
					}, 333);
				}

				function _wileOut() {
					$log.debug('wileOut');

					_mode = 'wile';

					_elements.bag.style.height = '';
					_elements.bag.style.opacity = '';
					_elements.footer.style.bottom = '';
					Array.prototype.slice.call(_elements.header.querySelectorAll('i.fa')).forEach(function(i) {
						i.style.visibility = '';
					});

					// adjust stage
					_elements.stage.style.height = '';

					$rootScope.$broadcast('chill-mode', false);
				}

				function _init() {
					var cancelOnIdleServiceIdle,
						cancelOnIdleServiceActive;

					cancelOnIdleServiceIdle = IdleService.onUserIdleFor(_chillModeActivateAfter, function() {
						// only activate chill mode under specific conditions
						if (scope.$parent.adminMode === false && _isFullScreen() === true && PlayerComponentRegistrar.getComponents().length > 0) {
							_chillOut();
						}
					});

					cancelOnIdleServiceActive = IdleService.onUserAction(function() {
						// only wile out if the user is in chill mode
						if (_mode === 'chill') {
							_wileOut();
						}
					});
					scope.$on('$destroy', function() {
						// clean up after ourselves
						cancelOnIdleServiceActive();
						cancelOnIdleServiceIdle();
					});
				}

				_init();
			}
		};
	}
}());
