(function() {
	'use strict';

	angular.module('ipconfigure.orchid.directive').directive('ipcDateTimePicker', ['$document', ipcDateTimePicker]);

	function ipcDateTimePicker($document) {
		return {
			scope: {
				model: '=ngModel',
				calendarId: '@'
			},
			templateUrl: 'views/date-time-picker.html',
			link: function(scope, element) {
				scope.dateTime = null;
				scope.opened = false;
				scope.lastEventTime = -1;
				scope.isDragging = false;

				$document.on('keydown', function(event) {
					if (event.keyCode === 27) {
						scope.closePicker();
					}
				});
				scope.closePicker = function($event) {

					// touch events are sometimes sent from iOS devices for reasons
					if (angular.isDefined($event)) {
						$event.stopPropagation();
						if ($event.type === "touchend") {
							// consume touch events that also generate a click
							if ($event.target.className.indexOf(scope.calendarId + "-toggle") > -1) {
								return;
							}
							// consume touches that end a drag
							else if (scope.isDragging) {
								scope.isDragging = false;
								return;
							}
						}
						else if ($event.type === "touchmove") {
							scope.isDragging = true;
							return;
						}
					}

					// This is awful, but to compensate for inconsistent event propagation across browsers, ensure that a different event is triggering us to close
					if (angular.isUndefined($event) || ($event.timeStamp !== scope.lastEventTime)) {
						scope.opened = false;
						var phase = this.$root.$$phase;
						if (phase !== '$apply') {
							scope.$apply();
						}
					}
				};
				scope.openPicker = function () {
					scope.dateTime = scope.model;
					scope.opened = true;
				};
				scope.togglePicker = function($event) {
					scope.lastEventTime = $event.timeStamp;
					if (scope.opened) {
						scope.closePicker();
					}
					else {
						scope.openPicker();
					}
				}
				scope.confirmPicker = function() {
					scope.opened = false;
					scope.model = scope.dateTime;
				}
		},
			require: 'ngModel',
			replace: true
		};
	}
}());
