(function() {
	'use strict';

	angular.module('ipconfigure.orchid.directive').directive('dynamicConfigWatcher', [dynamicConfigWatcher]);

	function dynamicConfigWatcher() {
		return {
			restrict: 'E',
			scope: {
				configuration: '=',
				capabilities: '='
			},
			link: function(scope) {
				function _hasCapability(caps, value) {
					return caps.options.some(function(o) {
						if (o === value) {
							return true;
						}
					});
				}

				function _watchConfig(which) {
					scope.$watch('configuration.' + which + '.Encoder', function(newValue) {
						var caps,
							key,
							curVal,
							min,
							max;

						if (newValue) {
							if (scope.capabilities[which] && scope.capabilities[which].Encoder.capabilities) {
								caps = scope.capabilities[which].Encoder.capabilities[newValue];

								// determine whether or not the current value valid for this newly selected capability, and change it to the default if not
								for (key in caps) {
									if (caps.hasOwnProperty(key) === true) {
										if (!scope.configuration[which][key]) {
											scope.configuration[which][key] = caps[key].default;
										}
										else {
											switch (caps[key].type) {
												case 'enum':
													if (_hasCapability(caps[key], scope.configuration[which][key]) === false) {
														scope.configuration[which][key] = caps[key].default;
													}
													break;
												case 'int':
													// only reset if outside min+max
													if (caps[key].hasOwnProperty('min') && caps[key].hasOwnProperty('max')) {
														min = parseInt(caps[key].min, 10);
														max = parseInt(caps[key].max, 10);
														curVal = parseInt(scope.configuration[which][key], 10);

														if (curVal > max || curVal < min) {
															scope.configuration[which][key] = caps[key].default;
														}
													}
													break;
												default:
													alert('type not supported yet', caps[key].type);
											}
										}
									}
								}
							}
						}
					});
				}

				_watchConfig('Audio');
				_watchConfig('Video');
			}
		};
	}
}());
