(function() {
	'use strict';

	angular.module('ipconfigure.orchid.directive').directive('dynamicInput', ['$compile', dynamicInput]);

	function dynamicInput($compile) {
		return {
			restrict: 'E',
			scope: {
				name: '=',
				value: '=',
				capability: '='
			},
			link: function(scope, element) {
				var html = '',
					attributes = '',
					minMaxStep = '';

				scope.normalizeValue = function() {
					var val = parseInt(scope.value, 10);

					if (val > scope.capability.max) {
						scope.value = scope.capability.max.toString();
					}
					else if (val < scope.capability.min) {
						scope.value = scope.capability.min.toString();
					}
				};

				if (scope.capability.type !== 'bool') {
					html += '<label for="' + scope.name + '">' + ipConfigure.core.splitCamelCase(scope.name) + ':</label>';
				}

				switch (scope.capability.type) {
					case 'string':
						html += '<input class="form-control" id="' + scope.name + '" type="text" ng-model="value" />';
						break;
					case 'ip_address':
						html += '<input class="form-control" id="' + scope.name + '" type="text" ng-model="value" placeholder="0.0.0.0"' + ((scope.capability.readOnly) ? 'disabled="true"' : '') + ' />';
						break;
					case 'int':
					case 'float':
						if (scope.capability.hasOwnProperty('min') && scope.capability.hasOwnProperty('max')) {
							if (scope.capability.min === scope.capability.max) {
								attributes = ' disabled="true"';
							}
							else {
								minMaxStep = ' min="' + scope.capability.min + '" max="' + scope.capability.max + '" ng-model="value" step="' + ((scope.capability.type === 'float') ? '0.1' : '1') + '"';
							}

							html += '<input id="' + scope.name + '" class="form-control" type="text" ng-model="value" class="input-mini"' + attributes + ' ng-blur="normalizeValue()" />';
							html += '<input type="range"' + minMaxStep + attributes + ' />';
						}
						else {
							html += '<input id="' + scope.name + '" class="form-control" type="number" ng-model="value" class="input-mini" />';
						}
						break;
					case 'enum':
						html += '<select class="form-control" id="' + scope.name + '" ng-model="value" ng-options="option for option in capability.options"></select>';
						break;
					case 'bool':
						html += '<label class="checkbox-inline" style="font-weight: bold;">';
						html += '<input type="checkbox" id="' + scope.name + '" ng-model="value">';
						html += scope.name;
						html += '</label>';
						break;
					default:
						html += '[unknown type: ' + scope.capability.type + ']';
						break;
				}

				element.append(html);
				$compile(element.contents())(scope);
			}
		};
	}
}());
