(function() {
	'use strict';

	angular.module('ipconfigure.orchid.directive').directive('scrollToOn', [scrollToOn]);

	function scrollToOn() {
		function _findScrollableParent(element) {
			var styles = window.getComputedStyle(element[0].parentNode, ''),
				overflowY = styles.getPropertyValue('overflow-y');

			if (overflowY === 'scroll') {
				return element.parent();
			}
			else {
				return _findScrollableParent(element.parent());
			}
		}

		function _scrollTopTween(scrollTop) {
			return function() {
				var i = d3.interpolateNumber(this.scrollTop, scrollTop);
				return function(t) {
					this.scrollTop = i(t);
				};
			};
		}

		return function(scope, element, attr) {
			scope.$on('scrollToOn', function(e, name) {
				var container,
					elementTop,
					containerTop;

				if (name === attr.scrollToOn) {
					container = _findScrollableParent(element);

					elementTop = element[0].getBoundingClientRect().top;
					containerTop = container.prop('offsetTop');

					// animate into view, taking into account the container's scrollTop
					d3.select(container[0])
						.transition()
						.duration(250)
						.tween('scrollTop', _scrollTopTween(elementTop - containerTop + container.prop('scrollTop')));
				}
			});
		};
	}
}());
