(function() {
	'use strict';

	angular.module('ipconfigure.orchid.directive').directive('stage', ['$log', '$rootScope', '$timeout', 'StageService', stage]);

	function stage($log, $rootScope, $timeout, StageService) {
		// z-index explanation
		//  - 83: normal cell
		//  - 85: cell being dragged
		//  - 87: cell just dragged/resized being animated over top of other cells
		//  - 101: full screen cell
		return {
			restrict: 'E',
			replace: true,
			scope: {},
			template: '<section></section>',
			link: function(scope, element) {
				var _container = d3.select(element[0]),
					_lineMatrixContainer = _container.append('svg').attr('class', 'line-matrix').style('width', '100%').style('height', '100%'),
					_cellMatrixContainer = _container.append('div').attr('class', 'cell-matrix'),
					_cellHighlightContainer = _container.append('svg').attr('class', 'cell-highlighter'),
					_backstagePadding = 20,
					_cellRatio = [16, 9],
					_maxCellPadding = 12,
					_cellPadding = 3,
					_duration = 200,
					_cellMatrix = {
						container: _cellMatrixContainer,
						columnsInsertedSinceRender: 0,  // values to assist in adding multiple columns at once
						rowsInsertedSinceRender: 0,     // values to assist in adding multiple columns at once
						cellWidth: 0,
						cellHeight: 0,
						offsetLeft: 0,
						offsetTop: 0,
						cells: [
							[_createCell()]
						]
					},
					_lineMatrix = {
						container: _lineMatrixContainer,
						containerSize: {
							width: 0,
							height: 0
						},
						firstColumnIndex: 1000,
						firstRowIndex: 1000,
						previousValues: {
							uninitialized: true,
							cellWidth: 0,
							cellHeight: 0,
							offsetLeft: 0,
							offsetTop: 0
						},
						horizontal: {
							group: _lineMatrixContainer.append('g').attr('orientation', 'horizontal'),
							lines: []
						},
						vertical: {
							group: _lineMatrixContainer.append('g').attr('orientation', 'vertical'),
							lines: []
						}
					},
					_skipAnimation = false,
					_aCellIsBeingAdded = false,
					_aCellIsFullScreen = false,
					_savedChillModeValues;

				// render functions

				function _calculateCellDimensions(rowCount, colCount) {
					var availableWidth = _lineMatrix.containerSize.width - (_backstagePadding * 2) - ((colCount - 1) * _cellPadding),
						availableHeight = _lineMatrix.containerSize.height - (_backstagePadding * 2) - ((rowCount - 1) * _cellPadding),
						availableCellWidth = availableWidth / colCount,
						availableCellHeight = availableHeight / rowCount,
						calculatedDimensions = {
							cellWidth: availableCellWidth,
							cellHeight: availableCellHeight,
							offsetLeft: 0,
							offsetTop: 0
						};

					// constrain aspect ratio
					if (availableCellWidth / _cellRatio[0] > availableCellHeight / _cellRatio[1]) {
						// calculate width from height and determine left offset
						calculatedDimensions.cellWidth = availableCellHeight / _cellRatio[1] * _cellRatio[0];
						calculatedDimensions.offsetLeft = (availableWidth - (calculatedDimensions.cellWidth * (colCount))) / 2;
					}
					else {
						// calculate height from width and determine top offset
						calculatedDimensions.cellHeight = availableCellWidth / _cellRatio[0] * _cellRatio[1];
						calculatedDimensions.offsetTop = (availableHeight - (calculatedDimensions.cellHeight * (rowCount))) / 2;
					}

					return calculatedDimensions;
				}

				function _render() {
					_lineMatrix.containerSize = {
						width: _container.node().clientWidth,
						height: _container.node().clientHeight
					};

					_renderCells();

					if (_skipAnimation === false) {
						_showLines();
					}

					_renderLines();

					if (_skipAnimation === false) {
						setTimeout(function() {
							_hideLines();
							_resizePlayers();
						}, _duration + 20);
					}

					_cellMatrix.columnsInsertedSinceRender = 0;
					_cellMatrix.rowsInsertedSinceRender = 0;
				}

				function _renderCells() {
					var colCount = _cellMatrix.cells[0].length,
						rowCount = _cellMatrix.cells.length,
						calculatedCellDimensions = _calculateCellDimensions(rowCount, colCount);

					_cellMatrix.cellWidth = calculatedCellDimensions.cellWidth;
					_cellMatrix.cellHeight = calculatedCellDimensions.cellHeight;
					_cellMatrix.offsetLeft = calculatedCellDimensions.offsetLeft;
					_cellMatrix.offsetTop = calculatedCellDimensions.offsetTop;

					// loop through the matrix and generate values for each cell
					_cellMatrix.cells.forEach(function(row, rowIndex) {
						row.forEach(function(cell, columnIndex) {
							var subject;

							cell.previousRow = cell.row;
							cell.previousColumn = cell.column;
							cell.row = rowIndex;
							cell.column = columnIndex;
							cell.left = (columnIndex * _cellMatrix.cellWidth) + (columnIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetLeft;
							cell.top = (rowIndex * _cellMatrix.cellHeight) + (rowIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetTop;
							// TODO too much ternary, fucker
							cell.width = cell.spannedUpon === true || cell.claimed !== true ? 0 : cell.columnSpan ? cell.columnSpan * _cellMatrix.cellWidth + (cell.columnSpan - 1) * _cellPadding : _cellMatrix.cellWidth;
							cell.height = cell.spannedUpon === true || cell.claimed !== true ? 0 : cell.rowSpan ? cell.rowSpan * _cellMatrix.cellHeight + (cell.rowSpan - 1) * _cellPadding : _cellMatrix.cellHeight;

							// create the element if it hasn't already been created
							_createElementForCell(cell);

							// set cell's initial values if defined
							if (cell.initialValues) {
								if (angular.isUndefined(cell.initialValues.left) === false) {
									cell.element.style('left', cell.initialValues.left + 'px');
								}

								if (angular.isUndefined(cell.initialValues.top) === false) {
									cell.element.style('top', cell.initialValues.top + 'px');
								}

								if (angular.isUndefined(cell.initialValues.width) === false) {
									cell.element.style('width', cell.initialValues.width + 'px');
								}

								if (angular.isUndefined(cell.initialValues.height) === false) {
									cell.element.style('height', cell.initialValues.height + 'px');
								}

								delete cell.initialValues;
							}

							// using transition() makes the operation async, so this conditional exists so that resize is handled smoothly
							if (_skipAnimation === false) {
								subject = cell.element.transition().duration(_duration);
							}
							else {
								subject = cell.element;
							}

							// move the cell to its destination
							subject
								.style('width', (cell.fullScreen === true ? cell.element.node().parentNode.parentNode.offsetWidth : cell.width) + 'px')
								.style('height', (cell.fullScreen === true ? cell.element.node().parentNode.parentNode.offsetHeight : cell.height) + 'px')
								.style('left', (cell.fullScreen === true ? 0 : cell.left) + 'px')
								.style('top', (cell.fullScreen === true ? 0 : cell.top) + 'px');

							// update the cell's data so it can be dragged
							cell.element.data([
								{ x: cell.left, y: cell.top }
							]);
						});
					});
				}

				function _renderLines() {
					var data;

					// initialize previous values so that we can base calculations on them
					if (_lineMatrix.previousValues.uninitialized === true) {
						delete _lineMatrix.previousValues.uninitialized;

						_lineMatrix.previousValues = {
							cellWidth: _cellMatrix.cellWidth,
							cellHeight: _cellMatrix.cellHeight,
							offsetLeft: _cellMatrix.offsetLeft,
							offsetTop: _cellMatrix.offsetTop
						};
					}

					if (_cellMatrix.cellWidth > 0 && _cellMatrix.cellHeight > 0) {
						_renderLinesGenerateHorizontal();
						_renderLinesGenerateVertical();

						data = _lineMatrix.horizontal.group.selectAll('line').data(_lineMatrix.horizontal.lines, function(d) { return d.index; });
						_renderLinesUpdating(data, 'y', _lineMatrix.containerSize.width);
						_renderLinesEntering(data, 'y', _lineMatrix.containerSize.width);
						_renderLinesExiting(data);

						data = _lineMatrix.vertical.group.selectAll('line').data(_lineMatrix.vertical.lines, function(d) { return d.index; });
						_renderLinesUpdating(data, 'x', _lineMatrix.containerSize.height);
						_renderLinesEntering(data, 'x', _lineMatrix.containerSize.height);
						_renderLinesExiting(data);
					}

					// save previous values for later calculations
					_lineMatrix.previousValues.cellWidth = _cellMatrix.cellWidth;
					_lineMatrix.previousValues.cellHeight = _cellMatrix.cellHeight;
					_lineMatrix.previousValues.offsetLeft = _cellMatrix.offsetLeft;
					_lineMatrix.previousValues.offsetTop = _cellMatrix.offsetTop;
				}

				function _createElementForCell(cell) {
					if (!cell.element) {
						cell.element = _cellMatrix.container.append('div')
//							.style('background-color', 'white')
							.style('width', cell.width + 'px')
							.style('height', cell.height + 'px');

						cell.element
							.on('click', function() { d3.event.preventDefault(); })
							.call(d3.behavior.drag()
								.origin(Object)
								.on('dragstart', (function(providedCell) {
									return function(d) {
										_onCellDragStart.call(this, d, providedCell);
									};
								}(cell)))
								.on('drag', (function(providedCell) {
									return function(d) {
										_onCellDrag.call(this, d, providedCell);
									};
								}(cell)))
								.on('dragend', (function(providedCell) {
									return function(d) {
										_onCellDragEnd.call(this, d, providedCell);
									};
								}(cell))));
					}
				}

				// TODO can these generate functions be combined? they are quite similar

				function _renderLinesGenerateHorizontal() {
					var oldLines = _lineMatrix.horizontal.lines,
						indexer,
						i,
						index,
						newY,
						initialY;

					_lineMatrix.horizontal.lines = [];

					// generates lines above first cell
					indexer = 0;

					for (i = _cellMatrix.offsetTop + _backstagePadding - _cellMatrix.cellHeight - _cellPadding; i > 0; i -= _cellMatrix.cellHeight + _cellPadding) {
						newY = i - _cellPadding / 2;
						index = _lineMatrix.firstRowIndex - 1 - indexer;

						if (oldLines.length === 0) {
							// if nothing has ever been rendered, then we don't need to calculate initialY, as there should be no animation
							initialY = newY;
						}
						else {
							// determine this line's ordinal position and then multiply it by the previous width to calculate where it would've appeared originally
							initialY = oldLines[0].value + (index - oldLines[0].index) * (_lineMatrix.previousValues.cellHeight + _cellPadding);
						}

						_lineMatrix.horizontal.lines.splice(0, 0, {
							index: index,
							value: newY,
							initialValue: initialY
						});

						indexer += 1;
					}

					// generate lines starting on the left edge of the first cell
					indexer = 0;

					for (i = _cellMatrix.offsetTop + _backstagePadding; i < _lineMatrix.containerSize.height; i += _cellMatrix.cellHeight + _cellPadding) {
						newY = i - _cellPadding / 2;
						index = _lineMatrix.firstRowIndex + indexer;

						if (oldLines.length === 0) {
							// if nothing has ever been rendered, then we don't need to calculate initialX, as there should be no animation
							initialY = newY;
						}
						else {
							initialY = 0;
							// combine the offset (where the first cell appears) with the previous cell width multiplied by the line's ordinal position to determine
							// where it would originally have appeared before the adjustment that was just made
							initialY = _lineMatrix.previousValues.offsetTop + _backstagePadding - _cellPadding / 2 + (index - _lineMatrix.firstRowIndex - _cellMatrix.rowsInsertedSinceRender) * (_lineMatrix.previousValues.cellHeight + _cellPadding);
						}

						_lineMatrix.horizontal.lines.push({
							index: index,
							value: newY,
							initialValue: initialY
						});

						indexer += 1;
					}
				}

				function _renderLinesGenerateVertical() {
					var oldLines = _lineMatrix.vertical.lines,
						indexer,
						i,
						index,
						newX,
						initialX;

					_lineMatrix.vertical.lines = [];

					// generates lines to the left of the the first cell
					indexer = 0;

					for (i = _cellMatrix.offsetLeft + _backstagePadding - _cellMatrix.cellWidth - _cellPadding; i > 0; i -= _cellMatrix.cellWidth + _cellPadding) {
						newX = i - _cellPadding / 2;
						index = _lineMatrix.firstColumnIndex - 1 - indexer;

						if (oldLines.length === 0) {
							// if nothing has ever been rendered, then we don't need to calculate initialX, as there should be no animation
							initialX = newX;
						}
						else {
							// determine this line's ordinal position and then multiply it by the previous width to calculate where it would've appeared originally
							initialX = oldLines[0].value + (index - oldLines[0].index) * (_lineMatrix.previousValues.cellWidth + _cellPadding);
						}

						_lineMatrix.vertical.lines.splice(0, 0, {
							index: index,
							value: newX,
							initialValue: initialX
						});

						indexer += 1;
					}

					// generate lines starting on the left edge of the first cell
					indexer = 0;

					for (i = _cellMatrix.offsetLeft + _backstagePadding; i < _lineMatrix.containerSize.width; i += _cellMatrix.cellWidth + _cellPadding) {
						newX = i - _cellPadding / 2;
						index = _lineMatrix.firstColumnIndex + indexer;

						if (oldLines.length === 0) {
							// if nothing has ever been rendered, then we don't need to calculate initialX, as there should be no animation
							initialX = newX;
						}
						else {
							initialX = 0;
							// combine the offset (where the first cell appears) with the previous cell width multiplied by the line's ordinal position to determine
							// where it would originally have appeared before the adjustment that was just made
//							console.log('ordinal', (index - _lineMatrix.firstColumnIndex - _cellMatrix.columnsInsertedSinceRender));
							initialX = _lineMatrix.previousValues.offsetLeft + _backstagePadding - _cellPadding / 2 + (index - _lineMatrix.firstColumnIndex - _cellMatrix.columnsInsertedSinceRender) * (_lineMatrix.previousValues.cellWidth + _cellPadding);
						}

						_lineMatrix.vertical.lines.push({
							index: index,
							value: newX,
							initialValue: initialX
						});

						indexer += 1;
					}
				}

				function _renderLinesUpdating(data, axis, max) {
					var otherAxis = axis === 'x' ? 'y' : 'x';

					data
						.attr('index', function(d) { return d.index; })
						.transition()
						.duration(_skipAnimation ? 0 : _duration)
						.style('stroke-width', _cellPadding)
						.attr(axis + '1', function(d) { return d.value; })
						.attr(axis + '2', function(d) { return d.value; })
						.attr(otherAxis + '1', 0)
						.attr(otherAxis + '2', max);
				}

				function _renderLinesEntering(data, axis, max) {
					var otherAxis = axis === 'x' ? 'y' : 'x';

					data.enter().append('line')
						.attr('index-orig', function(d) { return d.index; })
						.attr('index', function(d) { return d.index; })
						.style('stroke-width', _cellPadding)
						.attr(axis + '1', function(d) { return d.initialValue; })
						.attr(axis + '2', function(d) { return d.initialValue; })
						.transition()
						.duration(_skipAnimation ? 0 : _duration)
						.attr(axis + '1', function(d) { return d.value; })
						.attr(axis + '2', function(d) { return d.value; })
						.attr(otherAxis + '1', 0)
						.attr(otherAxis + '2', max);
				}

				function _renderLinesExiting(data) {
					// TODO can we animate this properly?
					// TODO we may need to calculate some sort of "finalValues" in the _delete functions, just like we've calculated "initialValues" in the insert/append functions
					data.exit()
//						.transition()
//						.duration(_skipAnimation ? 0 : _duration)
//						.attr(axis + '1', function (d) { return 0; })
//						.attr(axis + '2', function (d) { return 0; })
						.remove();
				}

				// create rows/columns

				function _createCell() {
					return {
						rowSpan: 1,
						columnSpan: 1,
						initialValues: {}
					};
				}

				function _insertRow() {
					var rowIndex,
						columnIndex,
						modifiedColumnIndex,
						cell;

					_lineMatrix.firstRowIndex -= 1;

					_cellMatrix.rowsInsertedSinceRender += 1;
					rowIndex = -(_cellMatrix.rowsInsertedSinceRender);

					_cellMatrix.cells.splice(0, 0, []);

					for (columnIndex = 0; columnIndex < _cellMatrix.cells[_cellMatrix.cells.length - 1].length; columnIndex += 1) {
						modifiedColumnIndex = columnIndex - _cellMatrix.columnsInsertedSinceRender;

						cell = _createCell();
						cell.initialValues.left = (modifiedColumnIndex * _cellMatrix.cellWidth) + (modifiedColumnIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetLeft;
						cell.initialValues.top = (rowIndex * _cellMatrix.cellHeight) + (rowIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetTop;
						cell.initialValues.width = _cellMatrix.cellWidth;
						cell.initialValues.height = _cellMatrix.cellHeight;

						_cellMatrix.cells[0].push(cell);
					}
				}

				function _insertColumn() {
					var rowIndex,
						columnIndex,
						modifiedRowIndex,
						cell;

					_lineMatrix.firstColumnIndex -= 1;

					_cellMatrix.columnsInsertedSinceRender += 1;
					columnIndex = -(_cellMatrix.columnsInsertedSinceRender);

					for (rowIndex = 0; rowIndex < _cellMatrix.cells.length; rowIndex += 1) {
						modifiedRowIndex = rowIndex - _cellMatrix.rowsInsertedSinceRender;

						cell = _createCell();
						cell.initialValues.left = (columnIndex * _cellMatrix.cellWidth) + (columnIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetLeft;
						cell.initialValues.top = (modifiedRowIndex * _cellMatrix.cellHeight) + (modifiedRowIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetTop;
						cell.initialValues.width = _cellMatrix.cellWidth;
						cell.initialValues.height = _cellMatrix.cellHeight;

						_cellMatrix.cells[rowIndex].splice(0, 0, cell);
					}
				}

				function _appendRow() {
					var columnIndex,
						rowIndex = _cellMatrix.cells.length,
						cell;

					_cellMatrix.cells.push([]);

					for (columnIndex = 0; columnIndex < _cellMatrix.cells[0].length; columnIndex += 1) {
						cell = _createCell();
						cell.initialValues.left = (columnIndex * _cellMatrix.cellWidth) + (columnIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetLeft;
						cell.initialValues.top = (rowIndex * _cellMatrix.cellHeight) + (rowIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetTop;
						cell.initialValues.width = _cellMatrix.cellWidth;
						cell.initialValues.height = _cellMatrix.cellHeight;

						_cellMatrix.cells[_cellMatrix.cells.length - 1].push(cell);
					}
				}

				function _appendColumn() {
					var rowIndex,
						columnIndex = _cellMatrix.cells[0].length,
						cell;

					for (rowIndex = 0; rowIndex < _cellMatrix.cells.length; rowIndex += 1) {
						cell = _createCell();
						cell.initialValues.left = (columnIndex * _cellMatrix.cellWidth) + (columnIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetLeft;
						cell.initialValues.top = (rowIndex * _cellMatrix.cellHeight) + (rowIndex * _cellPadding) + _backstagePadding + _cellMatrix.offsetTop;
						cell.initialValues.width = _cellMatrix.cellWidth;
						cell.initialValues.height = _cellMatrix.cellHeight;

						_cellMatrix.cells[rowIndex].push(cell);
					}
				}

				// delete rows/columns

				function _deleteFirstRow() {
					var i;

					if (_cellMatrix.cells.length > 1) {
						_lineMatrix.firstRowIndex += 1;
						_cellMatrix.rowsInsertedSinceRender -= 1;

						for (i = 0; i < _cellMatrix.cells[0].length; i += 1) {
							if (_cellMatrix.cells[0][i].element) {
								_cellMatrix.cells[0][i].element.remove();
							}
						}

						_cellMatrix.cells.splice(0, 1);
					}
				}

				function _deleteFirstColumn() {
					var i;

					if (_cellMatrix.cells[0].length > 1) {
						_cellMatrix.columnsInsertedSinceRender -= 1;
						_lineMatrix.firstColumnIndex += 1;

						for (i = 0; i < _cellMatrix.cells.length; i += 1) {
							if (_cellMatrix.cells[i][0].element) {
								_cellMatrix.cells[i][0].element.remove();
							}
							_cellMatrix.cells[i].splice(0, 1);
						}
					}
				}

				function _deleteLastRow() {
					var i;

					if (_cellMatrix.cells.length > 1) {
						for (i = 0; i < _cellMatrix.cells[_cellMatrix.cells.length - 1].length; i += 1) {
							if (_cellMatrix.cells[_cellMatrix.cells.length - 1][i].element) {
								_cellMatrix.cells[_cellMatrix.cells.length - 1][i].element.remove();
							}
						}

						_cellMatrix.cells.splice(_cellMatrix.cells.length - 1, 1);
					}
				}

				function _deleteLastColumn() {
					var i;

					if (_cellMatrix.cells[0].length > 1) {
						for (i = 0; i < _cellMatrix.cells.length; i += 1) {
							if (_cellMatrix.cells[i][_cellMatrix.cells[i].length - 1].element) {
								_cellMatrix.cells[i][_cellMatrix.cells[i].length - 1].element.remove();
							}
							_cellMatrix.cells[i].splice(_cellMatrix.cells[i].length - 1, 1);
						}
					}
				}

				function _removeEmptyRowsAndColumns() {
					var retVal = false;

					// top row empty?
					while (_cellMatrix.cells.length > 1 && _isRowEmpty(0) === true) {
						_deleteFirstRow();
						retVal = true;
					}

					// bottom row empty?
					while (_cellMatrix.cells.length > 1 && _isRowEmpty(_cellMatrix.cells.length - 1) === true) {
						_deleteLastRow();
						retVal = true;
					}

					// left column empty?
					while (_cellMatrix.cells[0].length > 1 && _isColumnEmpty(0) === true) {
						_deleteFirstColumn();
						retVal = true;
					}

					// right column empty?
					while (_cellMatrix.cells[0].length > 1 && _isColumnEmpty(_cellMatrix.cells[0].length - 1) === true) {
						_deleteLastColumn();
						retVal = true;
					}

					return retVal;
				}

				// cell move/resize

				function _onCellDragStart(d, cell) {
//					console.log('cell drag start');

					if (cell.claimed !== true) {
						return false;
					}

//					console.log('cell drag start', d3.event, d, cell);

//					// TODO only allow
//					if (cell.claimed !== true) {
//						console.log('dont allow');
//						d3.event.sourceEvent.stopPropagation();
//						return false;
//					}
				}

				function _onCellDrag(d, cell) {
					var relativeX,
						relativeY;

					if (cell.claimed !== true) {
						return false;
					}

					// chrome touch fix
					if (angular.isUndefined(d3.event.sourceEvent.clientX)) {
						if (window.TouchEvent && d3.event.sourceEvent instanceof window.TouchEvent) {
							d3.event.sourceEvent.clientX = d3.event.sourceEvent.changedTouches[0].clientX;
							d3.event.sourceEvent.clientY = d3.event.sourceEvent.changedTouches[0].clientY;
						}
					}

					if (cell.fullScreen !== true) {
						// initialize the drag
						if (angular.isUndefined(d.dragged)) {
							if (angular.isUndefined(d.initX)) {
								d.initX = d3.event.sourceEvent.clientX;
								d.initY = d3.event.sourceEvent.clientY;
							}
							else {
								// 2px threshold for detecting a drag
								if (Math.abs(d.initX - d3.event.sourceEvent.clientX) > 1 || Math.abs(d.initY - d3.event.sourceEvent.clientY) > 1) {
									relativeX = d.initX - cell.left - _container.node().offsetLeft;
									relativeY = d.initY - cell.top - _container.node().offsetTop;

									// figure out what they're trying to do
									if (relativeX >= cell.width - 21 && relativeY >= cell.height - 21) {
										// bottom right corner
										d.resizeOffsetX = cell.width - relativeX;
										d.resizeOffsetY = cell.height - relativeY;

										d.action = 'resize';
										d.resizeDirection = 'se';
									}
									else if (relativeX >= cell.width - 21 && cell.height / 2 - relativeY > -16 && cell.height / 2 - relativeY < 16) {
										// right edge
										d.resizeOffsetX = cell.width - relativeX;

										d.action = 'resize';
										d.resizeDirection = 'e';
									}
									else if (relativeY >= cell.height - 21 && cell.width / 2 - relativeX > -16 && cell.width / 2 - relativeX < 16) {
										// bottom edge
										d.resizeOffsetY = cell.height - relativeY;

										d.action = 'resize';
										d.resizeDirection = 's';
									}
									else {
										// if this cell has a span, determine which the user started dragging on
										if (cell.rowSpan) {
											d.rowOffset = Math.max(1, Math.ceil((relativeY) / (_cellMatrix.cellHeight + _cellPadding))) - 1;
										}
										else {
											d.rowOffset = 0;
										}

										if (cell.columnSpan) {
											d.columnOffset = Math.max(1, Math.ceil((relativeX) / (_cellMatrix.cellWidth + _cellPadding))) - 1;
										}
										else {
											d.columnOffset = 0;
										}

										d.action = 'move';
									}

									cell.element.style('opacity', 0.5);
									_showLines();

									d.dragged = true;
								}
							}
						}

						if (d.action === 'resize') {
							_onCellDragResize(d, cell);
						}
						else if (d.action === 'move') {
							_onCellDragMove(d, cell);
						}
					}
				}

				function _onCellDragMove(d, cell) {
					var dataSource = [],
						gridPosition,
						rootPosition,
						stagePosition,
						anchor,
						rowRange = {},
						columnRange = {},
						rowIndex,
						columnIndex;

					d.highlightColor = 'green';
					d.movePosition = undefined;

					//
					// STEP 1 | Swapping Cells: Is the mouse hovering over an existing cell?
					//
					d.moveTarget = scope.findCellInMatrixByCoordinates(d3.event.sourceEvent.clientX, d3.event.sourceEvent.clientY, cell);

					if (d.moveTarget) {
						dataSource.push({
							x: d.moveTarget.left,
							y: d.moveTarget.top,
							width: d.moveTarget.width,
							height: d.moveTarget.height
						});
					}
					else {
						//
						// STEP 2 | Expanding the Grid: Which grid coordinate are they dragging over?
						//
						gridPosition = scope.findGridPositionByCoordinates(d3.event.sourceEvent.clientX, d3.event.sourceEvent.clientY);

						// if the mouse isn't between cells (if it's not hovering over a grid line)
						if (gridPosition !== false) {
							stagePosition = {
								x: _container.node().offsetLeft,
								y: _container.node().offsetTop
							};

							anchor = {
								x: _lineMatrix.vertical.lines[0].value - _cellMatrix.cellWidth + stagePosition.x,
								y: _lineMatrix.horizontal.lines[0].value - _cellMatrix.cellHeight + stagePosition.y
							};

							if (cell.rowSpan) {
								rowRange.from = gridPosition[1] - d.rowOffset;
								rowRange.to = gridPosition[1] - d.rowOffset + cell.rowSpan - 1;
							}
							else {
								rowRange.from = gridPosition[1];
								rowRange.to = gridPosition[1];
							}

							if (cell.columnSpan) {
								columnRange.from = gridPosition[0] - d.columnOffset;
								columnRange.to = gridPosition[0] - d.columnOffset + cell.columnSpan - 1;
							}
							else {
								columnRange.from = gridPosition[0];
								columnRange.to = gridPosition[0];
							}

							rootPosition = _getRootGridPosition();

							for (rowIndex = rowRange.from; rowIndex <= rowRange.to; rowIndex += 1) {
								for (columnIndex = columnRange.from; columnIndex <= columnRange.to; columnIndex += 1) {
									dataSource.push(scope.getDimensionsForGridPosition(columnIndex, rowIndex, stagePosition, anchor));

									if (_cellMatrix.cells[rowIndex - rootPosition[1]] &&
										_cellMatrix.cells[rowIndex - rootPosition[1]][columnIndex - rootPosition[0]] &&
										_cellMatrix.cells[rowIndex - rootPosition[1]][columnIndex - rootPosition[0]].claimed === true &&
										_cellMatrix.cells[rowIndex - rootPosition[1]][columnIndex - rootPosition[0]] !== cell &&
										_cellMatrix.cells[rowIndex - rootPosition[1]][columnIndex - rootPosition[0]].spannedUponBy !== cell) {
										d.highlightColor = 'red';
									}
								}
							}

							// store destination position in d somewhere so dragEnd can use it and not have to recalculate
							d.movePosition = [gridPosition[0] - rootPosition[0] - d.columnOffset, gridPosition[1] - rootPosition[1] - d.rowOffset];
						}
					}

					scope.highlightCells(dataSource, d.highlightColor);

					d.dragged = true;
					d.x = d3.event.x;
					d.y = d3.event.y;

					cell.element
						.style('z-index', 85)
						.style('outline', '5px solid blue')
						.style('top', d.y + 'px')
						.style('left', d.x + 'px');
				}

				function _onCellDragResize(d, cell) {
					var rootPosition,
						cellPosition,
						spanPosition,
						desiredSpan,
						dataSource = [],
						highlightColor = 'green',
						ix,
						iy,
						cellToCheck;

					d.dragged = true;

					// sensible defaults for when we're stricken to 1 axis
					d.w = cell.width;
					d.h = cell.height;

					if (d.resizeDirection.indexOf('e') > -1) {
						// calculate height from mouse position
						d.w = d3.event.sourceEvent.clientX - cell.left - element[0].offsetLeft + d.resizeOffsetX;
					}

					if (d.resizeDirection.indexOf('s') > -1) {
						// calculate height from mouse position
						d.h = d3.event.sourceEvent.clientY - cell.top - element[0].offsetTop + d.resizeOffsetY;
					}

					rootPosition = _getRootGridPosition();
					cellPosition = scope.findGridPositionByCoordinates(cell.left + _cellPadding + _container.node().offsetLeft, cell.top + _cellPadding + _container.node().offsetTop);
					spanPosition = scope.findGridPositionByCoordinates(cell.left + d.w + _container.node().offsetLeft, cell.top + d.h + _container.node().offsetTop, true);
					desiredSpan = [spanPosition[0] - cellPosition[0] + 1, spanPosition[1] - cellPosition[1] + 1];

					if (desiredSpan[0] > 0 && desiredSpan[1] > 0) {
						// create dataSource
						for (iy = 0; iy < desiredSpan[1]; iy += 1) {
							for (ix = 0; ix < desiredSpan[0]; ix += 1) {
								dataSource.push(scope.getDimensionsForGridPosition(ix + cellPosition[0], iy + cellPosition[1]));

								// can we span here?
								if (_cellMatrix.cells[iy + cellPosition[1] - rootPosition[1]]) {
									cellToCheck = _cellMatrix.cells[iy + cellPosition[1] - rootPosition[1]][ix + cellPosition[0] - rootPosition[0]];

									if (cellToCheck && cellToCheck.claimed === true && cellToCheck !== cell && cellToCheck.spannedUponBy !== cell) {
										highlightColor = 'red';
									}
								}
							}
						}

						// highlight cells
						scope.highlightCells(dataSource, highlightColor);

						// set span
						if (highlightColor === 'red') {
							d.desiredSpan = [cell.columnSpan, cell.rowSpan];
						}
						else {
							d.desiredSpan = desiredSpan;
						}
					}
					else {
						// reset it back
						d.desiredSpan = [cell.columnSpan, cell.rowSpan];
					}

					// further constrain width and height to take the smaller of the two
//					if (d.resizeDirection === 'se') {
//						// constrain width and height to aspect ratio
//						constrainedWidth = d.h / _cellRatio[1] * _cellRatio[0];
//						constrainedHeight = d.w / _cellRatio[0] * _cellRatio[1];
//
//						if (constrainedHeight > d.h) {
//							d.w = d.h / _cellRatio[1] * _cellRatio[0];
//						}
//
//						if (constrainedWidth > d.w) {
//							d.h = d.w / _cellRatio[0] * _cellRatio[1];
//						}
//					}

					// determine how many cells this cell should span, favor width because 16x9
//					span = Math.ceil(d.w / (_cellMatrix.cellWidth + _cellPadding));

					// adjust the cell so it spans nicely across cells
					// i am not entirely sure why this calculation works as well as it does or why it's necesary, might be something to do w/ the _cellRatio
//					d.h += (span / 3) * _cellPadding;

//					// figure out if the cell can span this far
//					if (span === 0 || _isCellCapableOfSpanning(cell, span) === false) {
//						foundCellsColor = 'red';
//
//						// in case d.span existed in a previous event fire
//						delete d.span;
//					}
//					else {
//						d.span = span;
//					}
//
//					foundCells.push(_fillerCell);
//
//					_processSpannedCells(cell, span, function(spannedCell) {
//						foundCells.push(spannedCell);
//					});
//
//					_highlightCells(foundCells, foundCellsColor);

					// position and style the element
					cell.element
						.style('z-index', 85)
						.style('outline', '5px solid blue')
						.style('width', d.w + 'px')
						.style('height', d.h + 'px');
				}

				function _onCellDragEnd(d, cell) {
					if (cell.claimed !== true) {
						return false;
					}

					if (d.dragged === true) {
						// chrome touch fix
						if (angular.isUndefined(d3.event.sourceEvent.clientX)) {
							if (window.TouchEvent && d3.event.sourceEvent instanceof window.TouchEvent) {
								d3.event.sourceEvent.clientX = d3.event.sourceEvent.changedTouches[0].clientX;
								d3.event.sourceEvent.clientY = d3.event.sourceEvent.changedTouches[0].clientY;
							}
						}

//						_highlightCells([]);
//						_aCellIsBeingDragged = false;

						if (d.action === 'resize') {
							_onCellDragEndResize(d, cell);
						}
						else {
							_onCellDragEndMove(d, cell);
						}

						scope.persistStageConfiguration();

						_cellHighlightContainer.style('display', '');

						// clean up after ourselves so that the next drag action can fire cleanly
						delete d.action;
						delete d.dragged;
						delete d.initX;
						delete d.initY;
						delete d.resizeOffsetX;
						delete d.resizeOffsetY;
						delete d.rowSpanAnchor;
						delete d.columnSpanAnchor;
						delete d.moveTarget;
						delete d.rowOffset;
						delete d.columnOffset;
						delete d.highlightColor;
						delete d.w;
						delete d.h;
					}
				}

				function _onCellDragEndMove(d, cell) {
					var temp = {},
						adjustment;

					if (d.highlightColor === 'green' && d.moveTarget !== false) {
						// user swapped two cells

						// TODO swap cells in _cellMatrix.cells
						// TODO update rowSpan, columnSpan, and spannedUponBy
						// TODO animate them into place

						// clear old span information
						if (cell.rowSpan || cell.columnSpan) {
							_deleteCellSpanInformation(cell);
						}

						if (d.moveTarget.rowSpan || d.moveTarget.columnSpan) {
							_deleteCellSpanInformation(d.moveTarget);
						}

						// swap the source cell w/ the destination cell
						_cellMatrix.cells[d.moveTarget.row][d.moveTarget.column] = cell;
						_cellMatrix.cells[cell.row][cell.column] = d.moveTarget;

						// swap row/column since we're using them for calculations
						temp.row = cell.row;
						temp.column = cell.column;
						cell.row = d.moveTarget.row;
						cell.column = d.moveTarget.column;
						d.moveTarget.row = temp.row;
						d.moveTarget.column = temp.column;

						// swap span information
						temp.rowSpan = cell.rowSpan;
						temp.columnSpan = cell.columnSpan;
						cell.rowSpan = d.moveTarget.rowSpan;
						cell.columnSpan = d.moveTarget.columnSpan;
						d.moveTarget.rowSpan = temp.rowSpan;
						d.moveTarget.columnSpan = temp.columnSpan;

						// force the cell top+left to exactly where we dragged, so render() animates it correctly
						cell.top = d.y;
						cell.left = d.x;

						cell.element
							.style('z-index', 87)
							.style('opacity', 1);

						d.moveTarget.element.style('z-index', 85);

						// update new span information
						if (cell.rowSpan || cell.columnSpan) {
							_markCellsSpannedUpon(cell);
						}

						if (d.moveTarget.rowSpan || d.moveTarget.columnSpan) {
							_markCellsSpannedUpon(d.moveTarget);
						}

						// remove unclaimed cells
						_removeEmptyRowsAndColumns();

						// re-render
						_render();

						// TODO render should have a callback?
						temp = d.moveTarget;
						setTimeout(function() {
							cell.element
								.style('outline', 0)
								.style('z-index', '');

							temp.element.style('z-index', '');

							_resizePlayers();
						}, _duration);
					}
					else if (d.highlightColor === 'green' && Array.isArray(d.movePosition) === true && (cell.column !== d.movePosition[0] || cell.row !== d.movePosition[1])) {
						// user moved a cell to a new position, and we need to add rows/cols to account for it

						temp = cell;

						_cellMatrix.cells[temp.row][temp.column] = _createCell();
						_cellMatrix.cells[temp.row][temp.column].initialValues = {
							left: cell.left,
							top: cell.top,
							width: _cellMatrix.cellWidth,
							height: _cellMatrix.cellHeight
						};

						_deleteCellSpanInformation(cell);

						adjustment = _addCellsForNewPosition(d.movePosition, cell.columnSpan, cell.rowSpan);

						// simply move the cell over
						temp.column = d.movePosition[0] + adjustment[0];
						temp.row = d.movePosition[1] + adjustment[1];

						_cellMatrix.cells[d.movePosition[1] + adjustment[1]][d.movePosition[0] + adjustment[0]] = temp;

						temp.element
							.style('z-index', 87)
							.style('opacity', 1);

						_markCellsSpannedUpon(temp);

						// remove unclaimed cells
						_removeEmptyRowsAndColumns();

						_render();
						temp = cell;
						setTimeout(function() {
							cell.element
								.style('outline', 0)
								.style('z-index', '');

							temp.element.style('z-index', '');

							_resizePlayers();
						}, _duration);
					}
					else {
						// snap it back, because they tried dropping it somewhere they shouldn't have
						cell.element
							.style('opacity', '')
							.transition()
							.ease('linear')
							.duration(100)
							.style('width', cell.width + 'px')
							.style('height', cell.height + 'px')
							.style('top', cell.top + 'px')
							.style('left', cell.left + 'px')
							.each('end', function() {
								cell.element.style('z-index', '');
								cell.element.style('outline', 0);
								_resizePlayers();
							});

						d.x = cell.left;
						d.y = cell.top;

						// TODO eh?
						_lineMatrix.horizontal.group.style('opacity', 1).transition().duration(_duration).style('opacity', 0).each('end', function() {
							_hideLines();
							_lineMatrix.horizontal.group.style('opacity', 1);
						});
						_lineMatrix.vertical.group.style('opacity', 1).transition().duration(_duration).style('opacity', 0).each('end', function() {
							_hideLines();
							_lineMatrix.vertical.group.style('opacity', 1);
						});

//						setTimeout(function() {
//							cell.element.style('outline', 0);
//							_hideLines();
//						}, 120);
					}
				}

				function _onCellDragEndResize(d, cell) {
					if (d.desiredSpan[0] !== cell.columnSpan || d.desiredSpan[1] !== cell.rowSpan) {

						_deleteCellSpanInformation(cell);

						cell.columnSpan = d.desiredSpan[0];
						cell.rowSpan = d.desiredSpan[1];

						_addCellsForNewPosition([cell.column, cell.row], cell.columnSpan, cell.rowSpan);

						_markCellsSpannedUpon(cell);

						_removeEmptyRowsAndColumns();

						_render();

						// TODO render needs a callback
						setTimeout(function() {
							_resizePlayers();
						}, _duration);

						cell.element
							.style('opacity', '')
							.style('outline', 0)
							.style('z-index', '');
					}
					else {
						// snap it back, because they tried dropping it somewhere they shouldn't have
						cell.element
							.style('opacity', '')
							.style('outline', 0)
							.transition()
							.ease('linear')
							.duration(100)
							.style('width', cell.width + 'px')
							.style('height', cell.height + 'px')
							.style('top', cell.top + 'px')
							.style('left', cell.left + 'px')
							.each('end', function() {
								cell.element.style('z-index', '');
								_resizePlayers();
							});
					}

					setTimeout(function() {
						_hideLines();
					}, _duration);
				}

				// grid helper functions

				function _showLines() {
					_lineMatrix.horizontal.group.style('display', 'inline');
					_lineMatrix.vertical.group.style('display', 'inline');
				}

				function _hideLines() {
					_lineMatrix.horizontal.group.style('display', 'none');
					_lineMatrix.vertical.group.style('display', 'none');
				}

				function _isRowEmpty(rowNumber) {
					var columnCount = _cellMatrix.cells[rowNumber].length,
						i;

					for (i = 0; i < columnCount; i += 1) {
						if (_cellMatrix.cells[rowNumber][i].claimed === true || _cellMatrix.cells[rowNumber][i].spannedUpon === true) {
							return false;
						}
					}

					return true;
				}

				function _isColumnEmpty(columnNumber) {
					var rowCount = _cellMatrix.cells.length,
						i;

					for (i = 0; i < rowCount; i += 1) {
						if (_cellMatrix.cells[i][columnNumber].claimed === true || _cellMatrix.cells[i][columnNumber].spannedUpon === true) {
							return false;
						}
					}

					return true;
				}

				function _getRootGridPosition() {
					return scope.findGridPositionByCoordinates(_cellMatrix.cells[0][0].left + _container.node().offsetLeft + _cellPadding, _cellMatrix.cells[0][0].top + _container.node().offsetTop + _cellPadding);
				}

				function _processSpannedCellsByIndex(row, column, rowSpan, columnSpan, callback) {
					var iRow,
						iColumn,
						rowUpper,
						columnUpper;

					for (iRow = row, rowUpper = row + rowSpan; iRow < rowUpper; iRow += 1) {
						for (iColumn = column, columnUpper = column + columnSpan; iColumn < columnUpper; iColumn += 1) {
							if (!(iRow === row && iColumn === column)) {
								if (_cellMatrix.cells[iRow] && _cellMatrix.cells[iRow][iColumn] && callback(_cellMatrix.cells[iRow][iColumn]) === false) {
									return false;
								}
							}
						}
					}

					return true;
				}

				function _processSpannedCells(cell, rowSpan, columnSpan, callback) {
					return _processSpannedCellsByIndex(cell.row, cell.column, rowSpan, columnSpan, callback);
				}

				function _deleteCellSpanInformation(cell) {
					_processSpannedCells(cell, cell.rowSpan, cell.columnSpan, function(spannedUponCell) {
						delete spannedUponCell.claimed;
						delete spannedUponCell.spannedUpon;
						delete spannedUponCell.spannedUponBy;
						spannedUponCell.width = _cellMatrix.cellWidth;
						spannedUponCell.height = _cellMatrix.cellHeight;
						spannedUponCell.initialValues = {
							width: _cellMatrix.cellWidth,
							height: _cellMatrix.cellHeight
						};
					});
				}

				function _markCellsSpannedUpon(cell) {
					return _processSpannedCells(cell, cell.rowSpan, cell.columnSpan, function(spannedUponCell) {
						spannedUponCell.claimed = true;
						spannedUponCell.spannedUpon = true;
						spannedUponCell.spannedUponBy = cell;
						spannedUponCell.width = 0;
						spannedUponCell.height = 0;
						spannedUponCell.initialValues = {
							width: 0,
							height: 0
						};
						return true;
					});
				}

				function _addCellsForNewPosition(position, columnSpan, rowSpan) {
					var i,
						adjustment = [0, 0];

//					console.log('_addCellsForNewPosition', position, columnSpan, rowSpan);
//					console.log('matrix', _cellMatrix.cells[0].length, _cellMatrix.cells.length);

					// if the x position exceeds the width of the matrix, expand it
					if (position[0] + columnSpan - 1 >= _cellMatrix.cells[0].length) {
						for (i = _cellMatrix.cells[0].length; i <= position[0] + columnSpan - 1; i += 1) {
							_appendColumn();
						}
					}

					// if the y position exceeds the height of the matrix, expand it
					if (position[1] + rowSpan - 1 >= _cellMatrix.cells.length) {
						for (i = _cellMatrix.cells.length; i <= position[1] + rowSpan - 1; i += 1) {
							_appendRow();
						}
					}

					// if the x position is less than zero, insert some columns
					if (position[0] < 0) {
						for (i = 0; i > position[0]; i -= 1) {
							_insertColumn();

							adjustment[0] += 1;
						}
					}

					// if the y position is less than zero, insert some rows
					if (position[1] < 0) {
						for (i = 0; i > position[1]; i -= 1) {
							_insertRow();

							adjustment[1] += 1;
						}
					}

					return adjustment;
				}

				function _maximizeCell(cell) {
					var parentContainerWidth = cell.element.node().parentNode.parentNode.offsetWidth,
						parentContainerHeight = cell.element.node().parentNode.parentNode.offsetHeight;

					cell.element
						.style('z-index', 101)
						.transition()
						.style('top', '0px')
						.style('left', '0px')
						.style('width', (parentContainerWidth) + 'px')
						.style('height', (parentContainerHeight) + 'px')
						.each('end', function() {
							_resizePlayers(cell);
						});
				}

				function _restoreCell(cell) {
					// revert back from full screen
					cell.element.transition()
						.style('top', cell.top + 'px')
						.style('left', cell.left + 'px')
						.style('width', cell.width + 'px')
						.style('height', cell.height + 'px')
						.each('end', function() {
							cell.element.style('z-index', '');

							_resizePlayers(cell);
						});
				}

				function _resizePlayers(cell) {
					if (cell) {
						// if a cell was provided, just target it for resizing
						if (cell.element[0][0].children.length > 0) {
							angular.element(cell.element[0][0].children[0]).scope().$broadcast('resize.stage');
						}
					}
					else {
						scope.$broadcast('resize.stage');
					}
				}

				function _onKeyDown(e) {
					switch (e.which) {
						case 33: // page up, increase cell padding
							if (_cellPadding !== _maxCellPadding) {
								_cellPadding += 1;
								_render();
							}
							break;

						case 34: //page down, decrease cell padding
							if (_cellPadding !== 1) {
								_cellPadding -= 1;
								_render();
							}
							break;

//							case 37: // left, insert column or delete first column
//								if (e.ctrlKey === true) {
//									_deleteFirstColumn();
//									if (e.shiftKey === true) {
//										_deleteFirstColumn();
//									}
//								}
//								else {
//									_insertColumn();
//									if (e.shiftKey === true) {
//										_insertColumn();
//									}
//								}
//								_render();
//								break;
//
//							case 38: // up, insert row or delete first row
//								if (e.ctrlKey === true) {
//									_deleteFirstRow();
//									if (e.shiftKey === true) {
//										_deleteFirstRow();
//									}
//								}
//								else {
//									_insertRow();
//									if (e.shiftKey === true) {
//										_insertRow();
//									}
//								}
//								_render();
//								break;
//
//							case 39: // right, append column or delete last column
//								if (e.ctrlKey === true) {
//									_deleteLastColumn();
//									if (e.shiftKey === true) {
//										_deleteLastColumn();
//									}
//								}
//								else {
//									_appendColumn();
//									if (e.shiftKey === true) {
//										_appendColumn();
//									}
//								}
//								_render();
//								break;
//
//							case 40: // down, append row or delete last row
//								if (e.ctrlKey === true) {
//									_deleteLastRow();
//									if (e.shiftKey === true) {
//										_deleteLastRow();
//									}
//								}
//								else {
//									_appendRow();
//									if (e.shiftKey === true) {
//										_appendRow();
//									}
//								}
//								_render();
//								break;
//
//							case 107: // plus
//								if (e.altKey === false) {
//									_insertRow();
//									_insertColumn();
//									if (e.shiftKey === true) {
//										_insertRow();
//										_insertColumn();
//									}
//								}
//								else {
//									_appendRow();
//									_appendColumn();
//									if (e.shiftKey === true) {
//										_appendRow();
//										_appendColumn();
//									}
//								}
//								_render();
//								break;
//
//							case 109: // minus
//								break;
					}
				}

				// init
				function _init() {
					scope.$on('$destroy', function() {
						document.removeEventListener('keydown', _onKeyDown);
					});

					// $rootScope here to avoid infinite loop
					$rootScope.$on('resize', function(event, via) {
						if (via !== 'master-mode') {
							_skipAnimation = true;
						}

						_render();

						if (via !== 'master-mode') {
							_resizePlayers();
							_skipAnimation = false;
						}
					});

					document.addEventListener('keydown', _onKeyDown);

					$rootScope.$on('chill-mode', function(event, active) {
						if (active === true) {
							_savedChillModeValues = {
								cellPadding: _cellPadding,
								backstagePadding: _backstagePadding
							};

							_cellPadding = 0;
							_backstagePadding = 0;

							_render();
						}
						else {
							_cellPadding = _savedChillModeValues.cellPadding;
							_backstagePadding = _savedChillModeValues.backstagePadding;

							_savedChillModeValues = undefined;

							if (_aCellIsFullScreen === true) {
								$timeout(function() {
									_render();
								}, _duration + 1);
							}
							else {
								_render();
							}
						}
					});

					// ultimately, reloadStageConfiguration will be called and render the stage
					//_skipAnimation = true;
					//_render();
					//_skipAnimation = false;

					StageService.registerStage(scope);
				}

				// public methods

				scope.highlightCells = function(dataSource, color) {
					var data = _cellHighlightContainer.selectAll('rect').data(dataSource);

					data
						.attr('x', function(d) { return d.x + 'px'; })
						.attr('y', function(d) { return d.y + 'px'; })
						.attr('width', function(d) { return d.width + 'px'; })
						.attr('height', function(d) { return d.height + 'px'; })
						.style('fill', color);

					data.exit().remove();

					data.enter().append('rect')
						.attr('x', function(d) { return d.x + 'px'; })
						.attr('y', function(d) { return d.y + 'px'; })
						.attr('width', function(d) { return d.width + 'px'; })
						.attr('height', function(d) { return d.height + 'px'; })
						.style('fill', color)
						.style('opacity', 0.5);

					_cellHighlightContainer.style('display', dataSource.length === 0 ? '' : 'block');
				};

				scope.showLines = function() {
					_showLines();
				};

				scope.hideLines = function() {
					_hideLines();
				};

				scope.findCellInMatrixByCoordinates = function(x, y, ignoreCell) {
					var modifiedX = x - _container.node().offsetLeft,
						modifiedY = y - _container.node().offsetTop,
						foundCell = false;

					_cellMatrix.cells.some(function(row) {
						return row.some(function(cell) {
							if (cell !== ignoreCell && cell.left <= modifiedX && cell.left + cell.width >= modifiedX && cell.top <= modifiedY && cell.top + cell.height >= modifiedY) {
								foundCell = cell;
								return true;
							}
						});
					});

					return foundCell;
				};

				scope.findGridPositionByCoordinates = function(x, y, forceMath) {
					var stagePosition = { x: _container.node().offsetLeft, y: _container.node().offsetTop },
						anchorX = _lineMatrix.vertical.lines[0].value - _cellMatrix.cellWidth + stagePosition.x - _cellPadding / 2,
						anchorY = _lineMatrix.horizontal.lines[0].value - _cellMatrix.cellHeight + stagePosition.y - _cellPadding / 2,
						anchorMaxX = _lineMatrix.vertical.lines[_lineMatrix.vertical.lines.length - 1].value + stagePosition.x - _cellPadding / 2,
						anchorMaxY = _lineMatrix.horizontal.lines[_lineMatrix.horizontal.lines.length - 1].value + stagePosition.y - _cellPadding / 2,
						baseX,
						baseY,
						ix,
						iy,
						rowIndex = 0,
						columnIndex;

					if (forceMath === true || anchorX > x || anchorY > y || anchorMaxX < x || anchorMaxY < y) {
						// if either coordinate is out of bounds of the rendered grid, use math to determine the coordinate
						baseX = Math.floor((x - anchorX) / (_cellMatrix.cellWidth + _cellPadding));
						baseY = Math.floor((y - anchorY) / (_cellMatrix.cellHeight + _cellPadding));

						return [baseX, baseY];
					}
					else {
						// if the coordinates are over the rendered grid, let's loop so that the user cannot drop onto a visible grid line (because it "feels" better)
						for (iy = anchorY; iy <= _container.node().clientHeight + _cellMatrix.cellHeight; iy += _cellMatrix.cellHeight + _cellPadding) {
							columnIndex = 0;

							for (ix = anchorX; ix <= _container.node().clientWidth + _cellMatrix.cellWidth; ix += _cellMatrix.cellWidth + _cellPadding) {
								if (ix <= x && ix + _cellMatrix.cellWidth >= x && iy <= y && iy + _cellMatrix.cellHeight >= y) {
									return [columnIndex, rowIndex];
								}

								columnIndex += 1;
							}

							rowIndex += 1;
						}
					}

					return false;
				};

				scope.getDimensionsForGridPosition = function(x, y, stagePosition, anchor) {
					if (angular.isUndefined(stagePosition) === true) {
						stagePosition = {
							x: _container.node().offsetLeft,
							y: _container.node().offsetTop
						};
					}

					if (angular.isUndefined(anchor) === true) {
						anchor = {
							x: _lineMatrix.vertical.lines[0].value - _cellMatrix.cellWidth + stagePosition.x,
							y: _lineMatrix.horizontal.lines[0].value - _cellMatrix.cellHeight + stagePosition.y
						};
					}

					return {
						x: x * (_cellMatrix.cellWidth + _cellPadding) + anchor.x - stagePosition.x - _cellPadding / 2,
						y: y * (_cellMatrix.cellHeight + _cellPadding) + anchor.y - stagePosition.y - _cellPadding / 2,
						width: _cellMatrix.cellWidth,
						height: _cellMatrix.cellHeight
					};
				};

				scope.getCellByCoordinates = function(x, y) {
					return scope.findCellInMatrixByCoordinates(x, y);
				};

				scope.claimCellByCoordinates = function(x, y, callback) {
					var foundCell = scope.findCellInMatrixByCoordinates(x, y),
						gridPosition,
						rootPosition,
						adjustment;

					if (foundCell === false) {
						gridPosition = scope.findGridPositionByCoordinates(x, y);

						if (gridPosition !== false) {
							rootPosition = _getRootGridPosition();

							adjustment = _addCellsForNewPosition([gridPosition[0] - rootPosition[0], gridPosition[1] - rootPosition[1]], 1, 1);

							foundCell = _cellMatrix.cells[gridPosition[1] - rootPosition[1] + adjustment[1]][gridPosition[0] - rootPosition[0] + adjustment[0]];

							foundCell.claimed = true;

							_removeEmptyRowsAndColumns();
						}
					}

					if (foundCell === false) {
						// if the cell still doesn't exist (meaning they dropped it on a grid line), just fail the callback
						callback(false);
					}
					else {
						// officially claim the cell
						foundCell.claimed = true;

						// use pre-render so that the callback can animate along with render
						_render();
						callback(foundCell);
					}
				};

				scope.releaseCell = function(cell, preventRender) {
					var rowsOrColumnsDeleted = false;

					// TODO
//					if (cell.fullScreen === true) {
//						cell.fullScreen = false;
//						_aCellIsFullScreen = false;
//					}

					cell.claimed = false;

					_deleteCellSpanInformation(cell);

					cell.rowSpan = 1;
					cell.columnSpan = 1;

					if (preventRender !== true) {
						rowsOrColumnsDeleted = _removeEmptyRowsAndColumns();

						if (rowsOrColumnsDeleted === true && _cellMatrix.cells.length > 1 && _cellMatrix.cells[0].length > 1) {
							_showLines();
						}

						// TODO render callback
						_render();

						setTimeout(function() {
							_hideLines();
						}, _duration);
					}
				};

				// ~

				scope.obtainContainer = function(callback) {
					var iRow,
						iCol,
						rowLength = _cellMatrix.cells.length,
						columnLength = _cellMatrix.cells[0].length,
						newColumnDimensions,
						newRowDimensions,
						cell;

					if (_aCellIsBeingAdded === true) {
						return false;
					}

					// determine if an empty cell is available
					for (iRow = 0; iRow <= rowLength - 1; iRow += 1) {
						for (iCol = 0; iCol <= columnLength - 1; iCol += 1) {
							cell = _cellMatrix.cells[iRow][iCol];

							if (cell.claimed !== true && cell.spannedUpon !== true) {
								cell.claimed = true;

								// since unclaimed cells are 0x0, set up the width/height on first claim
								cell.width = _cellMatrix.cellWidth;
								cell.height = _cellMatrix.cellHeight;
								cell.element.style('width', cell.width + 'px');
								cell.element.style('height', cell.height + 'px');

								callback(cell);
								return true;
							}
						}
					}

					// there's no empty available cell, figure out if we need to add a row or a column
					_aCellIsBeingAdded = true;

					// this algorithm *attempts* to make the videos as large as possible by intelligently(?) expanding
					// in whatever direction will produce the largest cells

					// calculate width and height if a new column was  added and calculate total pixels
					newColumnDimensions = _calculateCellDimensions(rowLength, columnLength + 1);

					// and now if a new row was added
					newRowDimensions = _calculateCellDimensions(rowLength + 1, columnLength);

					if (newColumnDimensions.cellWidth * newColumnDimensions.cellHeight > newRowDimensions.cellWidth * newRowDimensions.cellHeight) {
						_appendColumn();

						cell = _cellMatrix.cells[0][_cellMatrix.cells[0].length - 1];
						cell.claimed = true;

						_render();
						callback(cell);

						setTimeout(function() {
							_aCellIsBeingAdded = false;
						}, _duration);

						return true;
					}
					else {
						_appendRow();

						cell = _cellMatrix.cells[_cellMatrix.cells.length - 1][0];
						cell.claimed = true;

						_render();

						callback(cell);

						setTimeout(function() {
							_aCellIsBeingAdded = false;
						}, _duration);

						return true;
					}
				};

				scope.toggleFullScreen = function(cell) {
					if (_aCellIsFullScreen === true) {
						if (cell.fullScreen === true) {
							_aCellIsFullScreen = false;
							cell.fullScreen = false;

							_restoreCell(cell);

							return true;
						}
						else {
							// sorry, a separate cell is already full screen
							return false;
						}
					}
					else {
						_aCellIsFullScreen = true;
						cell.fullScreen = true;

						_maximizeCell(cell);

						return true;
					}
				};

				scope.persistStageConfiguration = function() {
					var serializableMatrix = [],
						clonedRow;

					_cellMatrix.cells.forEach(function(row) {
						clonedRow = [];

						row.forEach(function(cell) {
							clonedRow.push(cell.claimed ? {
								streamId: cell.streamId,
								columnSpan: cell.columnSpan,
								rowSpan: cell.rowSpan
							} : {});
						});

						serializableMatrix.push(clonedRow);
					});

					localStorage.stageMatrix = JSON.stringify(serializableMatrix);
				};

				scope.reloadStageConfiguration = function(cellObtainedCallback) {
					if (localStorage.stageMatrix) {
						// delete originally created elements so we start with a clean slate
						_cellMatrix.cells.forEach(function(row) {
							row.forEach(function(cell) {
								if (cell.element) {
									cell.element.remove();
								}
							});
						});

						// load the stageMatrix from localStorage
						_cellMatrix.cells = JSON.parse(localStorage.stageMatrix);

						// set column/row properties so that we can span columns/rows correctly
						_cellMatrix.cells.forEach(function(row, rowIndex) {
							row.forEach(function(cell, columnIndex) {
								cell.column = columnIndex;
								cell.row = rowIndex;
							});
						});

						// create elements
						_cellMatrix.cells.forEach(function(row) {
							row.forEach(function(cell) {
								if (cell.streamId) {
									cell.claimed = true;

									// create an element so that the callback can stuff a player into it
									_createElementForCell(cell);

									// legacy, convert "span" to column/row span
									if (cell.span) {
										cell.columnSpan = cell.span;
										cell.rowSpan = cell.span;
										delete cell.span;
									}

									// deafult row/column span -- this is a hack fix... somehow cells were being persisted with no column/row span though
									if (angular.isUndefined(cell.columnSpan) === true) {
										cell.columnSpan = 1;
									}
									if (angular.isUndefined(cell.rowSpan) === true) {
										cell.rowSpan = 1;
									}

									// span the cells
									if (cell.columnSpan > 1 || cell.rowSpan > 1) {
										_markCellsSpannedUpon(cell);
									}

									cellObtainedCallback(cell.streamId, cell);
								}
							});
						});
					}

					_removeEmptyRowsAndColumns();

					_skipAnimation = true;
					_render();
					_skipAnimation = false;

					$log.debug('_matrix after reloadStageConfiguration', _cellMatrix.cells);
				};

				// initialize

				_init();
			}
		};
	}
}());
