(function() {
	'use strict';

	angular.module('ipconfigure.orchid.directive').directive('videoPreview', ['$log', '$timeout', videoPreview]);

	function _getRefreshDelay() {
		// between 30 and 90 seconds
		return Math.round(Math.random() * 30000 + 60000);
	}

	function videoPreview($log, $timeout) {
		return {
			restrict: 'E',
			replace: true,
			scope: {
				stream: '=',
				streamId: '=',
				autoPlay: '=?'
			},
			template: ' ' +
			'<div class="video-preview">' +
			'  <img>' +
			'  <img src="img/loading.gif" class="loading">' +
			'  <img src="img/stream-not-available.png">' +
			'  <img src="img/stream-failed.png">' +
			'</div>',
			link: function(scope, element) {
				var images = element.find('img'),
					previewImage = angular.element(images[0]),
					loadingImage = angular.element(images[1]),
					unavailableImage = angular.element(images[2]),
					failedImage = angular.element(images[3]),
					isVisible = false,
					lastResultSuccessful = false,
					refreshTimer;

				// bind to the img load event so we know when it's available to show
				previewImage.bind('load', function() {
					if (angular.isUndefined(scope.stream.aspectRatio) === true) {
						// forced delay because clientWidth/clientHeight are available immediately
						$timeout(function() {
							scope.stream.aspectRatio = [previewImage.prop('clientWidth'), previewImage.prop('clientHeight')];
						});
					}

					// hide/show the right bits
					loadingImage.css('display', 'none');
					previewImage.css('display', 'inline');
					unavailableImage.css('display', 'none');
					failedImage.css('display', 'none');

					lastResultSuccessful = Date.now();

					// refresh again after a specified delay
					$timeout.cancel(refreshTimer);
					refreshTimer = $timeout(scope.refresh, _getRefreshDelay());
				});

				// bind to the img load event so we know when to show the failed image
				previewImage.bind('error', function() {
					// hide/show the right bits
					loadingImage.css('display', 'none');
					previewImage.css('display', 'none');
					unavailableImage.css('display', 'inline');
					failedImage.css('display', 'none');

					lastResultSuccessful = false;

					// refresh again fairly quickly if an error occurred
					refreshTimer = $timeout(scope.refresh, 5000);
				});

				scope.$on('video-preview-play', function() {
					isVisible = true;
				});

				scope.$on('video-preview-stop', function() {
					isVisible = false;
				});

				scope.$on('$destroy', function() {
					$timeout.cancel(refreshTimer);
				});

				// this function will refresh the current frame under the following circumstances:
				//  - the last result was unsuccessful (or a frame has never been loaded)
				//  - the preview is actually visible (meaning the camera bag is out)
				//  - the last frame is stale (10 minutes old)
				scope.refresh = function(forceSource) {
					var src;

					if (forceSource === true || isVisible === true || lastResultSuccessful === false || (lastResultSuccessful + (1000 * 60 * 10)) < Date.now()) {
						src = scope.stream.href + '/frame?width=350&height=233&' + Date.now();

						$log.debug('view-preview pull frame', src);
						previewImage.attr('src', src);
					}
					else {
						// since the src isn't going to load or error out, we need to run the delay here
						$timeout.cancel(refreshTimer);
						refreshTimer = $timeout(scope.refresh, _getRefreshDelay());
					}
				};

				scope.$watch('stream', function() {
					_go();
				});

				scope.$watch('streamId', function() {
					// this was added because for some reason, the <video-preview> directive in admin-camera-edit.html will not
					// fire the 'stream' $watch here for some unknown reason, despite it most definitely changing.  $watching
					// streamId somehow fires the watch though, so I'm using it and and having to do special things because they
					// both fire initially.
					$timeout.cancel(refreshTimer);
					_go();
				});

				function _go() {
					// for some reason the first 'stream' we get is undefined
					if (angular.isUndefined(scope.stream) === false) {

						// hide/show the right bits
						loadingImage.css('display', 'inline');
						previewImage.css('display', 'none');
						unavailableImage.css('display', 'none');
						failedImage.css('display', 'none');

						// initial refresh
						scope.refresh(scope.autoPlay);
					}
				}
			}
		};
	}
}());
