(function() {
	'use strict';

	angular.module('ipconfigure.orchid.directive').directive('weekDaySchedule', ['$timeout', weekDaySchedule]);

	function weekDaySchedule($timeout) {
		function weekDayAxis() {
			var _scale,
				_width,
				_data;

			function weekDayAxisWorker(g) {
				var allowedSegments = [24 * 4 - 2, 24 * 2 - 2, 24 - 1, 12],
					widthPerSegment = 18,
					segmentCount = _width / widthPerSegment,
					i = 0,
					data;

				if (segmentCount === 0) {
					_data = [];
				}
				else {
					// determine the maximum segmentCount we can handle with the given width
					while (allowedSegments[i] > segmentCount) {
						i += 1;
					}

					segmentCount = allowedSegments[i];

					switch (segmentCount) {
						case 12:
							_data = d3.range(120, 1440, 120);
							break;
						case 23:
							_data = d3.range(60, 1440, 60);
							break;
						case 46:
							_data = d3.range(30, 1440, 30);
							break;
						case 94:
							_data = d3.range(15, 1440, 15);
							break;
					}
				}

				data = g.selectAll('text').data(_data);

				data.exit().remove();
				data.enter().append('text')
					.attr('y', 22);

				g.selectAll('text')
					.text(function(d) {
						switch (d % 60) {
							case 15:
							case 45:
								return ':';
							case 30:
								return '-';
							case 0:
								return ('00' + (d / 60)).slice(-2);
						}
					})
					.attr('x', function(d) {
						return _scale(d);
					});
			}

			weekDayAxisWorker.scale = function(v) {
				_scale = v;
				return weekDayAxisWorker;
			};

			weekDayAxisWorker.width = function(v) {
				_width = v;
				return weekDayAxisWorker;
			};

			return weekDayAxisWorker;
		}

		return {
			restrict: 'E',
			scope: {
				ngModel: '='
			},
			replace: true,
			template: ' ' +
				'<div class="week-day-schedule">' +
				'   <svg width="100%" height="37">' +
				'       <rect class="border" x="0" y="3" height="30" />' +
				'       <g class="axis"></g>' +
				'       <rect style="opacity: 0" width="100%" height="100%" />' +
				'       <g class="segments"></g>' +
				'       <rect class="new-segment" y="0" height="36" />' +
				'   </svg>' +
				'   <div class="toolbar">' +
				'       <div class="arrow-box">6:00 am - 1:00 pm</div>' +
				'       <div ng-click="deleteSegment()"><i class="fa fa-times"></i></div>' +
				'   </div>' +
				'</div>',
			link: function(scope, element) {
				var directiveId = Math.random(),
					svg = d3.select(element[0].firstElementChild),
					borderRect = svg.select('rect.border'),
					axisGroup = svg.select('g.axis'),
					segmentGroup = svg.select('g.segments'),
					newSegmentRect = svg.select('.new-segment'),
					toolbar = element[0].querySelector('.toolbar'),
					toolbarTimer,
					clientWidth,
					scale,
					weekDays,
					selectedSegment,
					dragInfo = {};

				function renderAxis() {
					clientWidth = element[0].clientWidth;

					scale = d3.scale.linear().range([0, clientWidth - 1]).domain([0, 1440]);
					weekDays = weekDayAxis().scale(scale).width(clientWidth - 1);

					axisGroup.call(weekDays);
				}

				function renderSegments() {
					var data,
						enterSelection,
						groupSelection,
						segmentDragBehavior;

					if (Array.isArray(scope.ngModel) === true) {
						data = segmentGroup.selectAll('g.segment').data(scope.ngModel);

						segmentDragBehavior = d3.behavior.drag()
							.origin(function(d) {
								return {
									x: scale(d.start)
								};
							})
							.on('dragstart', onSegmentDragStart)
							.on('drag', onSegmentDrag)
							.on('dragend', onSegmentDragEnd);

						// because of all the sub-grouping we're doing, we need to update children data
						segmentGroup.selectAll('g.segment g.drag-handle-left').data(scope.ngModel);
						segmentGroup.selectAll('g.segment g.drag-handle-right').data(scope.ngModel);
						segmentGroup.selectAll('g.segment rect.bg').data(scope.ngModel);

						data.exit().remove();

						enterSelection = data.enter().append('g').attr('class', 'segment')
							.on('touchstart', function(d) {
								toolbar.style.opacity = 1;
								positionToolbar(d);
							})
							.on('mouseenter', function(d) {
								$timeout.cancel(toolbarTimer);
								toolbar.style.opacity = 1;
								positionToolbar(d);
							})
							.on('mouseleave', function() {
								toolbarTimer = $timeout(function() {
									toolbar.style.opacity = 0;
								}, 1000);
							})
							.call(segmentDragBehavior);

						enterSelection.append('rect').attr('class', 'bg').attr('y', 0).attr('height', 36);

						groupSelection = enterSelection.append('g').attr('class', 'drag-handle-left');
						groupSelection.append('rect').attr('class', 'drag-handle').attr('x', 0).attr('y', 0).attr('width', 9).attr('height', 36).style('cursor', 'ew-resize');
						groupSelection.append('line').attr('x1', 3).attr('x2', 3).attr('y1', 4).attr('y2', 33).attr('stroke', '#999').style('cursor', 'ew-resize');
						groupSelection.append('line').attr('x1', 6).attr('x2', 6).attr('y1', 4).attr('y2', 33).attr('stroke', '#999').style('cursor', 'ew-resize');

						groupSelection = enterSelection.append('g').attr('class', 'drag-handle-right');
						groupSelection.append('rect').attr('class', 'drag-handle').attr('x', 0).attr('y', 0).attr('width', 9).attr('height', 36).style('cursor', 'ew-resize');
						groupSelection.append('line').attr('y1', 4).attr('y2', 33).attr('x1', 3).attr('x2', 3).attr('stroke', '#999').style('cursor', 'ew-resize');
						groupSelection.append('line').attr('y1', 4).attr('y2', 33).attr('x1', 6).attr('x2', 6).attr('stroke', '#999').style('cursor', 'ew-resize');

						segmentGroup.selectAll('g.segment').attr('transform', function(d) {
							return 'translate(' + scale(d.start) + ')';
						});

						segmentGroup.selectAll('g.drag-handle-right')
							.attr('transform', function(d) {
								var tranlateValue = Math.max(4.5, (scale(d.duration) - 9));
								return 'translate(' + tranlateValue + ')';
							});

						segmentGroup.selectAll('g rect.bg')
							.attr('width', function(d) {
								return scale(d.duration);
							});
					}
				}

				function positionElements() {
					borderRect.attr('width', clientWidth - 1);
				}

				function intToTime(i) {
					//var ampm = i < 1200 ? 'am' : 'pm';

					return ('00' + Math.floor(i / 60)).slice(-2) + ':' + ('00' + (i % 60)).slice(-2);// + ' ' + ampm;
				}

				function positionToolbar(d) {
					var time = intToTime(d.start) + ' - ' + intToTime(d.start + d.duration),
						toolbarLeft = Math.min(15 + scale(d.start), element[0].offsetLeft + clientWidth - toolbar.clientWidth);

					toolbar.style.left = toolbarLeft + 'px';
					toolbar.firstElementChild.textContent = time;

					selectedSegment = d;
				}

				function render() {
					renderAxis();
					renderSegments();
					positionElements();
				}

				function onSvgDragStart(d) {
					d.dragged = false;

					if (window.TouchEvent && d3.event.sourceEvent instanceof window.TouchEvent) {
						d3.event.sourceEvent.offsetX = d3.event.sourceEvent.changedTouches[0].clientX - element[0].getBoundingClientRect().left;
					}

					// firefox hack
					if (typeof d3.event.sourceEvent.offsetX === 'undefined') {
						d3.event.sourceEvent.offsetX = d3.event.sourceEvent.pageX - d3.event.sourceEvent.target.getBoundingClientRect().left;
					}

					// determine starting point
					d.initialStart = Math.round(scale.invert(d3.event.sourceEvent.offsetX) / 5) * 5;

					// determine bounds
					d.lowerBound = 0;
					d.upperBound = 1440;

					scope.ngModel.forEach(function(dd) {
						if (d.initialStart <= dd.start && dd.start < d.upperBound) {
							d.upperBound = dd.start + 1;
						}

						if (d.initialStart >= dd.start + dd.duration && dd.start + dd.duration > d.lowerBound) {
							d.lowerBound = dd.start + dd.duration;
						}
					});
				}

				function onSvgDrag(d) {
					var roundedStart = Math.round(scale.invert(d3.event.x) / 5) * 5;

					// initialize the drag
					if (d.dragged === false) {
						newSegmentRect.style('display', 'block');

						d.dragged = true;

						toolbar.style.opacity = 1;
					}

					// determine x and width
					if (d.initialStart > roundedStart) {
						d.start = roundedStart;
						d.duration = d.initialStart - roundedStart;
					}
					else {
						d.start = d.initialStart;
						d.duration = roundedStart - d.initialStart;
					}

					// restrain lower bounds
					if (d.lowerBound > d.start) {
						d.start = Math.max(d.lowerBound, d.start);
						d.duration = d.initialStart - d.lowerBound;
					}

					// restrain upper bounds
					if (d.upperBound < d.start + d.duration) {
						d.duration = d.upperBound - d.initialStart - 1;
					}

					// position the rect
					newSegmentRect.attr('x', scale(d.start));
					newSegmentRect.attr('width', scale(d.duration));

					positionToolbar(d);
				}

				function onSvgDragEnd(d) {
					var newSegment = {
						start: d.start,
						duration: d.duration
					};

					newSegmentRect.style('display', '');

					if (d.dragged === true && d.duration > 10) {
						selectedSegment = newSegment;

						scope.ngModel.push(newSegment);

						render();
						scope.$apply();
					}
					else {
						toolbar.style.opacity = 0;
					}
				}

				function onSegmentDragStart(d) {
					var d3Parent = d3.select(d3.event.sourceEvent.target.parentNode);

					d3.event.sourceEvent.stopPropagation();

					// store the rectangle we're dragging for later manipulation
					if (d3Parent.attr('class') !== 'segment') {
						dragInfo.draggingGroup = d3.select(d3.event.sourceEvent.target.parentNode.parentNode);
						dragInfo.dragType = d3Parent.attr('class').replace('drag-handle-', '');
					}
					else {
						dragInfo.draggingGroup = d3.select(d3.event.sourceEvent.target.parentNode);
						dragInfo.dragType = 'move';

					}

					// determine our dragging bounds by inspecting other segments after defining sensible defaults
					dragInfo.lowerBound = 0;
					dragInfo.upperBound = 1440;

					scope.ngModel.forEach(function(dd) {
						if (dd !== d && d.start <= dd.start && dd.start < dragInfo.upperBound) {
							dragInfo.upperBound = dd.start;
						}

						if (dd !== d && d.start >= dd.start + dd.duration && dd.start + dd.duration > dragInfo.lowerBound) {
							dragInfo.lowerBound = dd.start + dd.duration;
						}
					});
				}

				function onSegmentDrag(d) {
					var previousStart = d.start;

					toolbar.style.opacity = 1;

					switch (dragInfo.dragType) {
						case 'move':
							// update d while restricting the rectangle's bounds
							d.start = Math.min(dragInfo.upperBound - d.duration, Math.max(dragInfo.lowerBound, scale.invert(d3.event.x)));

							break;
						case 'left':
							// update d while restricting the rectangle's bounds
							d.start = Math.max(dragInfo.lowerBound, scale.invert(d3.event.x));
							d.duration -= d.start - previousStart;

							// limit how small of a selection they can make
							if (d.duration < 15) {
								d.start -= 15 - d.duration;
								d.duration = 15;
							}

							break;
						case 'right':
							if (angular.isUndefined(dragInfo.originalDuration) === true) {
								dragInfo.originalDuration = d.duration;
							}

							d.duration = Math.max(15, dragInfo.originalDuration - (d.start - scale.invert(d3.event.x)));

							if (d.start + d.duration > dragInfo.upperBound) {
								d.duration = dragInfo.upperBound - d.start;
							}

							break;
					}

					// round start to the nearest 5
					d.start = Math.round(d.start / 5) * 5;

					// round duration to the nearest 5
					d.duration = Math.round(d.duration / 5) * 5;

					// position the rect
					dragInfo.draggingGroup.attr('transform', 'translate(' + scale(d.start) + ')');
					dragInfo.draggingGroup.select('rect.bg').attr('width', scale(d.duration));
					dragInfo.draggingGroup.select('g.drag-handle-right')
						.attr('transform', 'translate(' + Math.max(4.5, (scale(d.duration) - 9)) + ')');

					// position the toolbar
					positionToolbar(d);
				}

				function onSegmentDragEnd() {
					dragInfo = {};
					scope.$apply();
				}

				function init() {
					// TODO slow your roll
					d3.select(window).on('resize.' + directiveId, render);

					d3.select(toolbar)
						.on('mouseenter', function() {
							$timeout.cancel(toolbarTimer);
						})
						.on('mouseleave', function() {
							toolbarTimer = $timeout(function() {
								toolbar.style.opacity = 0;
							}, 1000);
						});

					// drag events for creating new rectangles
					svg.data([
						{}
					]).call(d3.behavior.drag()
						.on('dragstart', onSvgDragStart)
						.on('drag', onSvgDrag)
						.on('dragend', onSvgDragEnd));

					scope.$watch('ngModel', function() {
						segmentGroup.selectAll('g.segment').data([]).exit().remove();
						renderSegments();
					});

					scope.$on('$destroy', function() {
						// unsubscribe from the resize event
						d3.select(window).on('resize.' + directiveId, null);
					});

					// initially render, the $watch will take care of renderSegments
					renderAxis();
					positionElements();
				}

				// without this $timeout, clientWidth wasn't taking into account scrollbars and was rendering incorrectly
				$timeout(init);

				scope.deleteSegment = function() {
					scope.ngModel.some(function(segment, index) {
						if (segment === selectedSegment) {
							// hide the toolbar w/out a fade
							toolbar.className = 'toolbar no-transition';
							toolbar.style.opacity = 0;
							$timeout(function() {
								toolbar.className = 'toolbar';
							});

							scope.ngModel.splice(index, 1);
							render();
							return true;
						}
					});
				};
			}
		};
	}
}());
