(function() {
	'use strict';

	angular.module('ipconfigure.orchid.service').factory('AuthService', ['$rootScope', '$http', '$log', '$q', 'BaseServiceUrl', AuthService]);

	function AuthService($rootScope, $http, $log, $q, BaseServiceUrl) {
		var authenticated = false,
			authenticatedUser;

		$rootScope.$on('renew-session', function(event, sessionInfo) {
			handleAuthenticationSuccess(sessionInfo);
		});

		function handleAuthenticationSuccess(response, deferred) {
			authenticated = true;
			authenticatedUser = {
				id: response.id,
				role: response.role,
				username: response.username,
				sessionId: response.sessionId
			};
			if(angular.isDefined(deferred)) {
				deferred.resolve();
			}
		}

		function handleAuthenticationFailure(data, status, deferred) {
			authenticated = false;
			authenticatedUser = undefined;

			if (status === 401) {
				// 401 is expected and will have a proper code/reason to pass back, otherwise we fabricate one
				deferred.reject(data);
			}
			else {
				$log.error('Unexpected response from auth service', data, status);

				deferred.reject({ code: status, reason: 'Unexpected response from auth service' });
			}
		}

		function authenticateViaCredentials(username, password, remember) {
			var deferred = $q.defer();

			$http.post(BaseServiceUrl + '/service/session', { username: username, password: password, remember: remember })
				.success(function(response) {
					handleAuthenticationSuccess(response, deferred);
				})
				.error(function(data, status) {
					handleAuthenticationFailure(data, status, deferred);
				});

			return deferred.promise;
		}

		function authenticateViaRememberCookie() {
			var deferred = $q.defer();

			$http.get(BaseServiceUrl + '/service/session')
				.success(function(response) {
					handleAuthenticationSuccess(response, deferred);
				})
				.error(function(data, status) {
					handleAuthenticationFailure(data, status, deferred);
				});

			return deferred.promise;
		}

		function getAuthenticatedUser() {
			return (authenticated) ? authenticatedUser : false;
		}

		function logOut() {
			var deferred = $q.defer();

			// tell service to delete the httpOnly cookies
			$http.delete(BaseServiceUrl + '/service/session')
				.success(function() {
					authenticated = false;
					authenticatedUser = undefined;

					deferred.resolve();
				})
				.error(function(data, status, headers, config) {
					$log.error('AuthService::logOut error', data, status, headers, config);

					deferred.reject();
				});

			return deferred.promise;
		}

		return {
			authenticateViaCredentials: authenticateViaCredentials,
			authenticateViaRememberCookie: authenticateViaRememberCookie,
			getAuthenticatedUser: getAuthenticatedUser,
			logOut: logOut
		};
	}
}());
