(function() {
	'use strict';

	angular.module('ipconfigure.orchid.service').factory('CameraService', ['$http', '$log', '$rootScope', 'OrchidDiscoveryService', CameraService]);

	function CameraService($http, $log, $rootScope, OrchidDiscoveryService) {
		var instanceLookup = {},
			cameras = [],
			state = { currentState: 'loading' };

		$rootScope.cameraServiceInstances = OrchidDiscoveryService.instances;

		// long-term we should use Object.observe here (polyfill or otherwise) instead of abusing rootScope
		$rootScope.$watchCollection('cameraServiceInstances', function(watchedInstances) {
			var i,
				key,
				newInstanceLookup = {};

			$log.debug('CameraService processing instances', watchedInstances);

			watchedInstances.forEach(function(instance) {
				// build our new instance lookup so we can easily keep track of added/removed instances
				newInstanceLookup[instance.id] = true;

				// is this a new instance?
				if (!instanceLookup.hasOwnProperty(instance.id)) {
					// TODO put this elsewhere in this service so it can be ran again to update a list of cameras for a given instance
					// grab a list of cameras for it
					$http.get(instance.baseUrl + '/service/cameras')
						.success(function(response) {
							var j;

							// we cannot simply overwrite cameras since it may be $watched
							if (angular.isUndefined(response) === false && Array.isArray(response.cameras) === true) {
								// add new entries
								for (j = 0; j < response.cameras.length; j += 1) {
									// link the instance to the camera
									response.cameras[j].instance = instance;

									// add the camera to our list
									cameras.push(response.cameras[j]);
								}
							}
							else {
								if (response.cameras !== null) {
									$log.error('unexpected response from /service/cameras', response);
								}
							}

							state.currentState = 'loaded';
						})
						.error(function(data, status, headers, config) {
							$log.error('request failed for /service/cameras', data, status, headers, config);

							state.currentState = 'loaded';
						});
				}
			});

			// iterate through our existing instances
			for (key in instanceLookup) {
				// has this instance been removed?
				if (newInstanceLookup.hasOwnProperty(key) === false) {
					// iterate through our cameras, removing any associated with the removed instance,
					for (i = 0; cameras.length > i; i += 1) {
						if (cameras[i].instance.id.toString() === key) {
							cameras.splice(i, 1);
							i -= 1;
						}
					}
				}
			}

			// TODO periodically grab new cameras? this will only help when multiple users are configuring cameras.

			// save the old lookup for later convenience
			instanceLookup = newInstanceLookup;
		});

		return {
			cameras: cameras,
			state: state
		};
	}
}());
