(function() {
	'use strict';

	angular.module('ipconfigure.orchid.service').factory('OrchidDiscoveryService', ['$http', '$log', '$timeout', 'AuthService', 'BaseServiceUrl', OrchidDiscoveryService]);

	function OrchidDiscoveryService($http, $log, $timeout, AuthService, BaseServiceUrl) {
		var instances = [],
			currentInstance = {},
			state = { currentState: 'loading' };

		function instanceExists(instanceToMatch, instanceList) {
			return instanceList.some(function(instance) {
				return instance.id === instanceToMatch.id;
			});
		}

		function mergeNewOrchidInstances(newInstances) {
			var i,
				key;

			// update currentInstance from newInstances
			for (i = 0; i < newInstances.length; i += 1) {
				if (newInstances[i].baseUrl === BaseServiceUrl) {
					// we're taking this approach because we can't simply overwrite currentInstance as it may be $watched
					for (key in newInstances[i]) {
						if (newInstances[i].hasOwnProperty(key) === true) {
							currentInstance[key] = newInstances[i][key];
						}
					}

					// since we had to new up a currentInstance, you won't be able to find currentInstance in instances by reference,
					// so we're overwiting this reference in instances with currentInstance
					newInstances[i] = currentInstance;
				}
			}

			// add new entries
			for (i = 0; i < newInstances.length; i += 1) {
				if (instanceExists(newInstances[i], instances) === false) {
					instances.push(newInstances[i]);
				}
			}

			// remove missing entries
			for (i = 0; i < instances.length; i += 1) {
				// looking for each of our current instances in the new response
				if (instanceExists(instances[i], newInstances) === false) {
					// if it doesn't exist anymore, remove it
					instances.splice(i, 1);
					i -= 1;
				}
			}
		}

		(function loop() {
			var retryIn = 333;

			if (AuthService.getAuthenticatedUser() !== false) {
				// grab discoverable orchids from the server
				$http.get(BaseServiceUrl + '/service/discoverable/orchids')
					.success(function(response) {
						if (angular.isUndefined(response) === false && Array.isArray(response.orchids) === true) {
							// merge instances from the server into our instances array. we cannot simply
							// overwrite instances since it may be $watched.
							mergeNewOrchidInstances(response.orchids);
						}
						else {
							$log.error('unexpected response from /service/discoverable/orchids', response);
						}
					})
					.error(function(data, status, headers, config) {
						$log.error('request failed for /discoverable/orchids', data, status, headers, config);
					});

				// delay our re-query a good while, because discoverable orchids shouldn't change all that often
				retryIn = 1000 * 60 * 60;
			}

			$timeout(loop, retryIn);
		}());

		function updateLicenseSession(licenseSession) {
			var key;

			// copy up response.license
			for (key in licenseSession.license) {
				if (licenseSession.license.hasOwnProperty(key) === true) {
					currentInstance.licenseSession.license[key] = licenseSession.license[key];
				}
			}

			// delete removed properties (for retention mainly)
			for (key in currentInstance.licenseSession.license) {
				if (licenseSession.license.hasOwnProperty(key) === false) {
					delete currentInstance.licenseSession.license[key];
				}
			}

			// copy up response.state
			for (key in licenseSession.state) {
				if (licenseSession.state.hasOwnProperty(key) === true) {
					currentInstance.licenseSession.state[key] = licenseSession.state[key];
				}
			}
		}

		return {
			currentInstance: currentInstance,
			instances: instances,
			state: state,
			updateLicenseSession: updateLicenseSession
		};
	}
}());
