(function() {
	'use strict';

	angular.module('ipconfigure.orchid.service').factory('StageService', [StageService]);

	function StageService() {
		var _stage,
			_aCellIsFullScreen = false;

		return {
			registerStage: function(stage) {
				// this is not the most Angular way of doing things, but for the problem at hand, it seems necessary
				_stage = stage;
			},
			dragClear: function() {
				_stage.showLines();
				_stage.highlightCells([]);
			},
			dragOver: function(x, y) {
				var cell,
					gridPosition,
					dataSource = [];

				_stage.showLines();

				cell = _stage.findCellInMatrixByCoordinates(x, y);

				if (cell !== false) {
					// highlight the cell
					dataSource.push({
						x: cell.left,
						y: cell.top,
						width: cell.width,
						height: cell.height
					});
				}
				else {
					gridPosition = _stage.findGridPositionByCoordinates(x, y);

					if (gridPosition !== false) {
						dataSource.push(_stage.getDimensionsForGridPosition(gridPosition[0], gridPosition[1]));
					}
				}

				// highlight the coordinates
				_stage.highlightCells(dataSource, 'green');
			},
			drop: function(hideLines) {
				if (hideLines === true) {
					_stage.hideLines();
				}

				_stage.highlightCells([]);
			},
			isACellFullScreen: function() {
				return _aCellIsFullScreen;
			},
			obtainContainer: function(callback) {
				_stage.obtainContainer(callback);
			},
			getCellByCoordinates: function(x, y) {
				return _stage.getCellByCoordinates(x, y);
			},
			claimCellByCoordinates: function(x, y, callback) {
				_stage.claimCellByCoordinates(x, y, callback);
			},
			releaseCell: function(cell, preventRender) {
				_stage.releaseCell(cell, preventRender);
			},
			toggleFullScreen: function(cell) {
				_stage.toggleFullScreen(cell);

				_aCellIsFullScreen = cell.fullScreen;
			},
			persistStageConfiguration: function() {
				_stage.persistStageConfiguration();
			},
			reloadStageConfiguration: function(cellObtainedCallback) {
				_stage.reloadStageConfiguration(cellObtainedCallback);
			}
		};
	}
}());
