(function() {
	'use strict';

	angular.module('ipconfigure.orchid.service').factory('StreamService', ['$log', '$rootScope', 'CameraService', StreamService]);

	function StreamService($log, $rootScope, CameraService) {
		var streams = [],
			state = { currentState: 'loading' };

		$rootScope.streamService = { cameras: CameraService.cameras, cameraServiceState: CameraService.state };

		// long-term we should use Object.observe here (polyfill or otherwise) instead of abusing rootScope
		$rootScope.$watchCollection('streamService.cameras', forceUpdate);

		$rootScope.$watchCollection('streamService.cameraServiceState.currentState', function(watchedState) {
			// since the StreamService doesn't make any $http calls itself, we rely on the CameraService's state
			state.currentState = watchedState;
		});

		function forceUpdate() {
			$log.debug('StreamService processing cameras', $rootScope.streamService.cameras);

			// empty streams out
			while (streams.length > 0) {
				streams.splice(0, 1);
			}

			// add all primary streams
			$rootScope.streamService.cameras.forEach(function(camera) {
				var filteredStreams;

				if (Array.isArray(camera.streams) === true) {
					filteredStreams = camera.streams.filter(function(stream) {
						return stream.id === camera.primaryStream.id;
					});

					if (filteredStreams.length > 0) {
						// link the instance to the camera
						filteredStreams[0].cameraExpanded = camera;

						// generate the rtsp live url
						filteredStreams[0].liveStream = camera.instance.rtspProtocol + '://' + ipConfigure.core.parseUri(camera.instance.baseUrl).host + ((camera.instance.rtspPort) ? (':' + camera.instance.rtspPort) : '') + '/orchid/live?id=' + filteredStreams[0].id;

						// generate the thumb url
						filteredStreams[0].thumb = filteredStreams[0].href + '/frame';

						// TODO is this still relevant? scope.$broadcast('video-preview-play');

						// add the stream to our list
						streams.push(filteredStreams[0]);
					}
				}
			});
		}

		return {
			streams: streams,
			state: state,
			forceUpdate: forceUpdate
		};
	}
}());
